-- -------------------------------------------------------------------------- --
-- FlightMaster by kunda                                                      --
-- -------------------------------------------------------------------------- --
--                                                                            --
-- FlightMaster shows all available flight points.                            --
-- FlightMaster is a flight path time-remaining addon.                        --
--                                                                            --
-- Features:                                                                  --
-- # FlightMaster is language independent!                                    --
--   The node data is saved in a LOCALE INDEPENDENT way.                      --
--   You can start WoW with any language and you always get the same data.    --
-- # Contains all available Taxi Nodes.                                       --
-- # Contains all Taxi Node NPC-IDs (deci from hexa UnitGUID() [pos 6-10]).   --
--   (This allows to check them on different websites)                        --
-- # Contains required condition(s) to access a flight point (if necessary).  --
-- # Can show missing (and unreachable) flight points in TaxiRouteMap.        --
-- # Can correct the "TaxiButton"..i - FrameLevel bug in TaxiRouteMap.        --
-- # Shows a frame while on a flight with duration and other info.            --
--   - Three different world views: 'Normal', 'Energy' and 'Panorama'.        --
--     -> The 'Energy' option (Energy Saving Mode) reduces energy consumption --
--        by approximately 20%-35% (depends). That's true, test it!           --
-- # Flight duration:                                                         --
--   - More than 2500 pre-defined flight duration data.                       --
--   - Flight duration estimate for unknown routes is done in 4 steps:        --
--     1. other faction OR                                                    --
--     2. reversed OR                                                         --
--     3. other faction reversed OR                                           --
--     4. estimate: single hops AND/OR longest known routes (like 1. - 3.)    --
--   - Guild Perk Level 21 (MoP) reduces the flight duration by 25%.          --
--   - Actions that stops flight-time recording:                              --
--     - Summon:       function ConfirmSummon                                 --
--     - Battleground: function AcceptBattlefieldPort                         --
--     - World PvP:    function BattlefieldMgrEntryInviteResponse             --
--     - LFG:          event LFG_UPDATE                                       --
--     - Pet Battle:   event PET_BATTLE_QUEUE_STATUS                          --
--  -> The saved data is always the basic duration (without guild bonus) and  --
--     is always without any flown-times-divided-through average calculation. --
-- # Accessible via LibDataBroker addon, Minimap Button, Interface Options or --
--   slash command '/flightmaster' or '/fm'.                                  --
--                                                                            --
-- Notes:                                                                     --
-- # The following 'flight points' are NOT included because they are one-way  --
--   flights and the standard taxi events/functions does not work:            --
--   - The Druid-only flight points (Alliance and Horde) in Nighthaven.       --
--   - The two Sha'tari Skyguard flight points in Outland.                    --
-- # Acherus is a valid flight point, but only available for Death Knights.   --
--                                                                            --
-- Credits:                                                                   --
-- The 'flight duration estimate'-code for unknown routes contains parts from --
-- FuBar_ToFu from Kemayo.                                                    --
--                                                                            --
-- -------------------------------------------------------------------------- --

local _, _, _, tocversion = GetBuildInfo()
if tocversion < 50400 then
	print("|cffffff82FlightMaster:|r", "Pandaria FlightPoint remap with Patch 5.4.0.17205.")
	print("|cffffff82FlightMaster:|r", "This version is for tocversion 50400 or newer only!")
	return
end

-- ---------------------------------------------------------------------------------------------------------------------
FlightMaster_Data = {}        -- SavedVariables - FlightTimes
FlightMaster_Nodes = {}       -- SavedVariables - localized flight points
FlightMaster_Options = {}     -- SavedVariables - options
FlightMaster_PlayerNodes = {} -- SavedVariablesPerCharacter known flight points

local FlightMaster = CreateFrame("Frame")     -- container
local L = FlightMaster_Locales                -- localization table

local FlightMaster_NPC     = FlightMaster_NPC     -- NPC DATA
local FlightMaster_exNPC   = FlightMaster_exNPC   -- NPC exception DATA
local FlightMaster_DEFAULT = FlightMaster_DEFAULT -- default flight duration data

local FlightMaster_local_Nodes = {} -- local FlightMaster_Nodes -> current locale only: InitOptions()

local myTAXI_MAP_WIDTH = TAXI_MAP_WIDTH
local myTAXI_MAP_HEIGHT = TAXI_MAP_HEIGHT
local myTAXI_BUTTON_MIN_DIST = TAXI_BUTTON_MIN_DIST

local GVAR = {}     -- UI Widgets
local TEMPLATE = {} -- Templates

local playerFaction
local oppositeFaction

local AddonIcon = "Interface\\TaxiFrame\\UI-Taxi-Icon-Green"

local locale = GetLocale()

local inWorld
local inTaxi
local mapError
local flightMap
local flightBonus
local flightStart
local flightDestination
local flightDuration = 0
local flightEstimate = ""
local flightHops = 0
local flightCost = 0
local flightNodes = {}
local flightNone = {} -- known flight points, but unreachable
local flightSteps = {}
local reposNode = {}
--local flightKnown = {} -- TEST only

local GUI_Data = {}        -- data table
local GUI_Data_Count = 0   -- data table count
local GUI_SortBy = 3       -- data table sort by
local GUI_SortOrder = 1    -- data table sort order
local GUI_Data_maxNum = 50 -- data table max rows
local GUI_Data_minNum = 5  -- data table min rows
local GUI_Data_curNum = 35 -- data table current rows
local GUI_Selection
local SearchQuery          -- search

local mult4 = 10^4

local _G = _G

local type  = type
local pairs = pairs

local math_fmod  = math.fmod
local math_floor = math.floor

local string_format = string.format
local string_find   = string.find
local string_lower  = string.lower
local string_match  = string.match

local table_sort = table.sort
local table_wipe = table.wipe

local GetTime  = GetTime
local UnitGUID = UnitGUID

local NumTaxiNodes       = NumTaxiNodes
local GetNumRoutes       = GetNumRoutes
local TaxiNodePosition   = TaxiNodePosition
local TaxiNodeName       = TaxiNodeName
local TaxiNodeGetType    = TaxiNodeGetType
local TaxiNodeCost       = TaxiNodeCost
local TaxiNodeSetCurrent = TaxiNodeSetCurrent

local GetMoney       = GetMoney
local GetMoneyString = GetMoneyString

local SetMapToCurrentZone    = SetMapToCurrentZone
local GetCurrentMapContinent = GetCurrentMapContinent

local TaxiGetSrcX  = TaxiGetSrcX
local TaxiGetSrcY  = TaxiGetSrcY
local TaxiGetDestX = TaxiGetDestX
local TaxiGetDestY = TaxiGetDestY

local offset = 3

local m1, m2, m3, m4, m5, m6 = GetMapContinents()
local continentID = {
	[1] = {nameID = "Kalimdor",        name = m1 or "Kalimdor",        icon = "Interface\\Icons\\Achievement_Zone_Kalimdor_01"},
	[2] = {nameID = "EasternKingdoms", name = m2 or "EasternKingdoms", icon = "Interface\\Icons\\Achievement_Zone_EasternKingdoms_01"},
	[3] = {nameID = "Outland",         name = m3 or "Outland",         icon = "Interface\\Icons\\Achievement_Zone_Outland_01"},
	[4] = {nameID = "Northrend",       name = m4 or "Northrend",       icon = "Interface\\Icons\\Achievement_Zone_Northrend_01"},
	[5] = {nameID = "Pandaria",        name = m6 or "Pandaria",        icon = "Interface\\Icons\\ExpansionIcon_MistsofPandaria"},
}

local TaxiErrors = {
	[ERR_TAXINOPATH] = 1,
	[ERR_TAXINOPATHS] = 1,
	[ERR_TAXINOSUCHPATH] = 1,
	[ERR_TAXINOTELIGIBLE] = 1,
	[ERR_TAXINOTENOUGHMONEY] = 1,
	[ERR_TAXINOTSTANDING] = 1,
	[ERR_TAXINOTVISITED] = 1,
	[ERR_TAXINOVENDORNEARBY] = 1,
	[ERR_TAXIPLAYERALREADYMOUNTED] = 1,
	[ERR_TAXIPLAYERBUSY] = 1,
	[ERR_TAXIPLAYERMOVING] = 1,
	[ERR_TAXIPLAYERSHAPESHIFTED] = 1,
	[ERR_TAXISAMENODE] = 1,
	[ERR_TAXITOOFARAWAY] = 1,
	[ERR_TAXIUNSPECIFIEDSERVERERROR] = 1,
}

local heightFontNormal = math_floor( select(2, _G["GameFontNormal"]:GetFont()) +0.5 )
local heightFontSmall  = math_floor( select(2, _G["GameFontNormalSmall"]:GetFont()) +0.5 )

local width_default = 420
local width_SelectionButtonSpace = 10
local width_DataList_Number = 30
local width_DataList_NPC = 60
local width_DataList_Exceptions = 55

local height_baseheadline = 28
local height_headline = 28-1
local height_baseselection = 5+heightFontNormal+5+32+5+17+5+heightFontSmall+5
local height_selection = height_baseselection -1
local height_data = offset+(1*12)+2+(GUI_Data_curNum*12)+offset
local height_basestatus = 20
local height_status = 20-1
local height_dataMin = offset+(1*12)+2+(GUI_Data_minNum*12)+offset
local height_dataMax = offset+(1*12)+2+(GUI_Data_maxNum*12)+offset
local height_mainframe = height_headline + height_selection + height_data + height_status
local height_mainframeMin = height_headline + height_selection + height_dataMin + height_status
local height_mainframeMax = height_headline + height_selection + height_dataMax + height_status

local barWidth = 400
local barHeight = 5+ heightFontNormal+2+heightFontNormal+2+heightFontNormal+20+16 +5

local worldPoints, worldNumPoints, worldHeight, worldWidth = {}, 0, 0, 0
local worldSound
local barEffScale, barScale
-- ---------------------------------------------------------------------------------------------------------------------

-- ---------------------------------------------------------------------------------------------------------------------
local function Print(...)
	print("|cffffff82FlightMaster:|r", ...)
end

local function calcID(x, y)
	-- 1606192 number 0.60551518201828 0.19244165718555 Gadgetzan, Tanaris
	local nx = (math_floor((x*1000)+0.5)*1000) + 1000000
	local ny = math_floor((y*1000)+0.5)
	local n = nx+ny
	--print(n, nx, ny, x, y)
	return n
end

local function formatTime(duration)
	if duration > 0 then
		return string_format("%d:%02d", duration/60, math_fmod(duration, 60))
	else
		return "?:??"
	end
end

local function currentTimeDate(form)
	local YYYY, MM, DD = date("%Y"), date("%m"), date("%d")
	local hour, minute = GetGameTime()
	if hour   < 10 then hour   = "0"..hour   end
	if minute < 10 then minute = "0"..minute end
	    if form == 1 then return tonumber(YYYY..MM..DD..hour..minute)                 -- form 1: 201112312359         number
	elseif form == 2 then return tostring(YYYY..MM..DD..hour..minute)                 -- form 2: "201112312359"       string
	elseif form == 3 then return YYYY.."-"..MM.."-"..DD.."---"..hour.."-"..minute     -- form 3: "2011-12-31---23-59" string
	elseif form == 4 then return YYYY.."-"..MM.."-"..DD.."-"..hour.."-"..minute       -- form 4: "2011-12-31-23-59"   string
	else                  return YYYY.."-"..MM.."-"..DD.." | "..hour..":"..minute end -- form 5: "2011-12-31 | 23:59" string
end

local months = {MONTH_JANUARY,MONTH_FEBRUARY,MONTH_MARCH,MONTH_APRIL,MONTH_MAY,MONTH_JUNE,MONTH_JULY,MONTH_AUGUST,MONTH_SEPTEMBER,MONTH_OCTOBER,MONTH_NOVEMBER,MONTH_DECEMBER}
local function localizedTimeDate(input, form)
	local YYYY, MM, DD, hour, minute = strmatch(tostring(input), "(%d%d%d%d)%D-(%d%d)%D-(%d%d)%D-(%d%d)%D-(%d%d)") -- input: 2011 * 12 * 31 * 23 * 59
	YYYY = (YYYY or "????") MM = (months[tonumber(MM)] or "??") DD = (DD or "??") hour = (hour or "??") minute = (minute or "??")
	if form == 1 then return hour..":"..minute.." | "..DD.."-"..MM.."-"..YYYY     -- form1: 23:59 | 31-December-2011
	else              return YYYY.."-"..MM.."-"..DD.." | "..hour..":"..minute end -- form2: 2011-December-31 | 23:59
end
-- ---------------------------------------------------------------------------------------------------------------------

-- ---------------------------------------------------------------------------------------------------------------------
local function NOOP() end

local function Desaturation(texture, desaturation)
	local shaderSupported = texture:SetDesaturated(desaturation)
	if not shaderSupported then
		if desaturation then
			texture:SetVertexColor(0.5, 0.5, 0.5)
		else
			texture:SetVertexColor(1.0, 1.0, 1.0)
		end
	end
end
-- ---------------------------------------------------------------------------------------------------------------------

-- ---------------------------------------------------------------------------------------------------------------------
-- Template BorderTRBL START ----------------------------------------
TEMPLATE.BorderTRBL = function(frame, color) -- TRBL = Top-Right-Bottom-Left
	frame.FrameBorder = frame:CreateTexture(nil, "BORDER")
	frame.FrameBorder:SetPoint("TOPLEFT", 1, -1)
	frame.FrameBorder:SetPoint("BOTTOMRIGHT", -1, 1)
	frame.FrameBorder:SetTexture(0, 0, 0, 1)
	frame.FrameBackground = frame:CreateTexture(nil, "BACKGROUND")
	frame.FrameBackground:SetPoint("TOPLEFT", 0, 0)
	frame.FrameBackground:SetPoint("BOTTOMRIGHT", 0, 0)
	frame.FrameBackground:SetTexture(0.4, 0.4, 0.4, 1)
end
-- Template BorderTRBL END ----------------------------------------

-- Template TextButton START ----------------------------------------
TEMPLATE.DisableTextButton = function(button)
	button.Border:SetTexture(0.4, 0.4, 0.4, 1)
	button:Disable()
end

TEMPLATE.EnableTextButton = function(button, action)
	local buttoncolor
	if action == 1 then
		bordercolor = {0.73, 0.26, 0.21, 1}
	elseif action == 2 then
		bordercolor = {0.43, 0.32, 0.68, 1}
	elseif action == 3 then
		bordercolor = {0.24, 0.46, 0.21, 1}
	elseif action == 4 then
		bordercolor = {0.73, 0.26, 0.21, 1}
	else
		bordercolor = {1, 1, 1, 1}
	end
	button.Border:SetTexture(bordercolor[1], bordercolor[2], bordercolor[3], bordercolor[4])
	button:Enable()
end

TEMPLATE.TextButton = function(button, text, action)
	local buttoncolor
	local bordercolor
	if action == 1 then
		button:SetNormalFontObject("GameFontNormal")
		button:SetDisabledFontObject("GameFontDisable") 
		buttoncolor = {0.38, 0, 0, 1}
		bordercolor = {0.73, 0.26, 0.21, 1}
	elseif action == 2 then
		button:SetNormalFontObject("GameFontNormalSmall")
		button:SetDisabledFontObject("GameFontDisableSmall")
		buttoncolor = {0, 0, 0.5, 1}
		bordercolor = {0.43, 0.32, 0.68, 1}
	elseif action == 3 then
		button:SetNormalFontObject("GameFontNormalSmall")
		button:SetDisabledFontObject("GameFontDisableSmall")
		buttoncolor = {0, 0.2, 0, 1}
		bordercolor = {0.24, 0.46, 0.21, 1}
	elseif action == 4 then
		button:SetNormalFontObject("GameFontNormalSmall")
		button:SetDisabledFontObject("GameFontDisableSmall")
		buttoncolor = {0.38, 0, 0, 1}
		bordercolor = {0.73, 0.26, 0.21, 1}
	else
		button:SetNormalFontObject("GameFontNormal")
		button:SetDisabledFontObject("GameFontDisable")
		buttoncolor = {0, 0, 0, 1}
		bordercolor = {1, 1, 1, 1}
	end

	button.Background = button:CreateTexture(nil, "BORDER")
	button.Background:SetPoint("TOPLEFT", 1, -1)
	button.Background:SetPoint("BOTTOMRIGHT", -1, 1)
	button.Background:SetTexture(0, 0, 0, 1)

	button.Border = button:CreateTexture(nil, "BACKGROUND")
	button.Border:SetPoint("TOPLEFT", 0, 0)
	button.Border:SetPoint("BOTTOMRIGHT", 0, 0)
	button.Border:SetTexture(bordercolor[1], bordercolor[2], bordercolor[3], bordercolor[4])

	button.Normal = button:CreateTexture(nil, "ARTWORK")
	button.Normal:SetPoint("TOPLEFT", 2, -2)
	button.Normal:SetPoint("BOTTOMRIGHT", -2, 2)
	button.Normal:SetTexture(buttoncolor[1], buttoncolor[2], buttoncolor[3], buttoncolor[4])
	button:SetNormalTexture(button.Normal)

	button.Disabled = button:CreateTexture(nil, "OVERLAY")
	button.Disabled:SetPoint("TOPLEFT", 3, -3)
	button.Disabled:SetPoint("BOTTOMRIGHT", -3, 3)
	button.Disabled:SetTexture(0.6, 0.6, 0.6, 0.2)
	button:SetDisabledTexture(button.Disabled)

	button.Highlight = button:CreateTexture(nil, "OVERLAY")
	button.Highlight:SetPoint("TOPLEFT", 3, -3)
	button.Highlight:SetPoint("BOTTOMRIGHT", -3, 3)
	button.Highlight:SetTexture(0.6, 0.6, 0.6, 0.2)
	button:SetHighlightTexture(button.Highlight)

	button:SetPushedTextOffset(1, -1)	
	button:SetText(text)
end
-- Template TextButton END ----------------------------------------

-- Template IconButton START ----------------------------------------
TEMPLATE.IconButton = function(button, cut)
	button.Back = button:CreateTexture(nil, "BORDER")
	button.Back:SetPoint("TOPLEFT", 1, -1)
	button.Back:SetPoint("BOTTOMRIGHT", -1, 1)
	button.Back:SetTexture(0, 0, 0, 1)

	button.Border = button:CreateTexture(nil, "BACKGROUND")
	button.Border:SetPoint("TOPLEFT", 0, 0)
	button.Border:SetPoint("BOTTOMRIGHT", 0, 0)
	button.Border:SetTexture(0.4, 0.4, 0.4, 1)

	if cut == 1 then
		button.Normal = button:CreateTexture(nil, "ARTWORK")
		button.Normal:SetPoint("TOPLEFT", 3, -3)
		button.Normal:SetPoint("BOTTOMRIGHT", -3, 3)
		button.Normal:SetTexture("Interface\\Buttons\\UI-Panel-MinimizeButton-Up")
		button.Normal:SetTexCoord(8/32, 22/32, 10/32, 22/32)
		button:SetNormalTexture(button.Normal)
		button.Push = button:CreateTexture(nil, "ARTWORK")
		button.Push:SetPoint("TOPLEFT", 4, -4)
		button.Push:SetPoint("BOTTOMRIGHT", -4, 4)
		button.Push:SetTexture("Interface\\Buttons\\UI-Panel-MinimizeButton-Up")
		button.Push:SetTexCoord(8/32, 22/32, 10/32, 22/32)
		button:SetPushedTexture(button.Push)
	elseif cut == 2 then
		button.Normal = button:CreateTexture(nil, "ARTWORK")
		button.Normal:SetPoint("TOPLEFT", 3, -3)
		button.Normal:SetPoint("BOTTOMRIGHT", -3, 3)
		button.Normal:SetTexture("Interface\\GossipFrame\\BinderGossipIcon")
		button:SetNormalTexture(button.Normal)
		button.Push = button:CreateTexture(nil, "ARTWORK")
		button.Push:SetPoint("TOPLEFT", 4, -4)
		button.Push:SetPoint("BOTTOMRIGHT", -4, 4)
		button.Push:SetTexture("Interface\\GossipFrame\\BinderGossipIcon")
		button:SetPushedTexture(button.Push)
	end

	button.Highlight = button:CreateTexture(nil, "OVERLAY")
	button.Highlight:SetPoint("TOPLEFT", 3, -3)
	button.Highlight:SetPoint("BOTTOMRIGHT", -3, 3)
	button.Highlight:SetTexture(0.6, 0.6, 0.6, 0.2)
	button:SetHighlightTexture(button.Highlight)
end
-- Template IconButton END ----------------------------------------

-- Template CheckButton START ----------------------------------------
TEMPLATE.DisableCheckButton = function(button)
	button.Text:SetTextColor(0.5, 0.5, 0.5)
	button.Border:SetTexture(0.4, 0.4, 0.4, 1)
	button:Disable()
end

TEMPLATE.EnableCheckButton = function(button)
	button.Text:SetTextColor(1, 1, 1)
	button.Border:SetTexture(0.8, 0.2, 0.2, 1)
	button:Enable()
end

TEMPLATE.CheckButton = function(button, size, space, text, textcolor)
	button.Border = button:CreateTexture(nil, "BACKGROUND")
	button.Border:SetWidth( size )
	button.Border:SetHeight( size )
	button.Border:SetPoint("LEFT", 0, 0)
	button.Border:SetTexture(0.4, 0.4, 0.4, 1)

	button.Background = button:CreateTexture(nil, "BORDER")
	button.Background:SetPoint("TOPLEFT", button.Border, "TOPLEFT", 1, -1)
	button.Background:SetPoint("BOTTOMRIGHT", button.Border, "BOTTOMRIGHT", -1, 1)
	button.Background:SetTexture(0, 0, 0, 1)

	button.Normal = button:CreateTexture(nil, "ARTWORK")
	button.Normal:SetPoint("TOPLEFT", button.Border, "TOPLEFT", 1, -1)
	button.Normal:SetPoint("BOTTOMRIGHT", button.Border, "BOTTOMRIGHT", -1, 1)
	button.Normal:SetTexture(0, 0, 0, 1)
	button:SetNormalTexture(button.Normal)

	button.Push = button:CreateTexture(nil, "ARTWORK")
	button.Push:SetPoint("TOPLEFT", button.Border, "TOPLEFT", 4, -4)
	button.Push:SetPoint("BOTTOMRIGHT", button.Border, "BOTTOMRIGHT", -4, 4)
	button.Push:SetTexture(0.4, 0.4, 0.4, 0.5)
	button:SetPushedTexture(button.Push)

	button.Disabled = button:CreateTexture(nil, "ARTWORK")
	button.Disabled:SetPoint("TOPLEFT", button.Border, "TOPLEFT", 3, -3)
	button.Disabled:SetPoint("BOTTOMRIGHT", button.Border, "BOTTOMRIGHT", -3, 3)
	button.Disabled:SetTexture(0.4, 0.4, 0.4, 0.5)
	button:SetDisabledTexture(button.Disabled)

	button.Checked = button:CreateTexture(nil, "ARTWORK")
	button.Checked:SetWidth( size )
	button.Checked:SetHeight( size )
	button.Checked:SetPoint("LEFT", 0, 0)
	button.Checked:SetTexture("Interface\\Buttons\\UI-CheckBox-Check")
	button:SetCheckedTexture(button.Checked)

	button.Text = button:CreateFontString(nil, "OVERLAY", "GameFontNormalSmall")
	button.Text:SetHeight( size )
	button.Text:SetPoint("LEFT", button.Normal, "RIGHT", space, 0)
	button.Text:SetJustifyH("LEFT")
	button.Text:SetText(text)
	if textcolor then
		button.Text:SetTextColor(textcolor[1], textcolor[2], textcolor[3], 1)
	else
		button.Text:SetTextColor(1, 1, 1, 1)
	end

	button:SetWidth(size + space + button.Text:GetStringWidth() + space)
	button:SetHeight(size)

	button.Highlight = button:CreateTexture(nil, "OVERLAY")
	button.Highlight:SetPoint("TOPLEFT", button, "TOPLEFT", 0, 0)
	button.Highlight:SetPoint("BOTTOMRIGHT", button, "BOTTOMRIGHT", 0, 0)
	button.Highlight:SetTexture(1, 1, 1, 0.1)
	button.Highlight:Hide()

	button:SetScript("OnEnter", function() button.Highlight:Show() end)
	button:SetScript("OnLeave", function() button.Highlight:Hide() end)
end
-- Template CheckButton END ----------------------------------------

-- Template EditBox START ----------------------------------------
TEMPLATE.EditBox = function(button, width, showIcon, funcSearch, funcClear)
	local fxWidth = 17
	if not showIcon then
		fxWidth = 0.00000001
	end
	local boxwidth = width-(fxWidth+2+2+17)

	button.SearchIcon = button:CreateTexture(nil, "ARTWORK")
	button.SearchIcon:SetWidth(fxWidth)
	button.SearchIcon:SetHeight(fxWidth)
	button.SearchIcon:SetPoint("LEFT", 0, 0)
	button.SearchIcon:SetTexture("Interface\\Minimap\\Tracking\\None")
	button.SearchIcon:SetTexCoord(3/32, 31/32, 1/32, 29/32)

	button.EditBox = CreateFrame("EditBox", nil, button)
	button.EditBox:SetWidth(boxwidth)
	button.EditBox:SetHeight(17)
	button.EditBox:SetPoint("LEFT", button.SearchIcon, "RIGHT", 2, 0)
	button.EditBox:SetAutoFocus(false)
	button.EditBox:SetFontObject(GameFontNormalSmall)
	button.EditBox:SetTextColor(1, 1, 1, 1)
	button.EditBox:SetTextInsets(2, 0, 0, 0)

	button.EditBoxLeft = button.EditBox:CreateTexture(nil, "BACKGROUND")
	button.EditBoxLeft:SetWidth(5)
	button.EditBoxLeft:SetHeight(17)
	button.EditBoxLeft:SetPoint("LEFT", 0, 0)
	button.EditBoxLeft:SetTexture("Interface\\Chatframe\\UI-ChatInputBorderFocus-Left")
	button.EditBoxLeft:SetTexCoord(8/32, 13/32, 8/32, 25/32)
	button.EditBoxLeft:SetVertexColor(0.6, 0.6, 0.6, 1)

	button.EditBoxMiddle = button.EditBox:CreateTexture(nil, "BACKGROUND")
	button.EditBoxMiddle:SetWidth(boxwidth-5-5)
	button.EditBoxMiddle:SetHeight(17)
	button.EditBoxMiddle:SetPoint("LEFT", button.EditBoxLeft, "RIGHT", 0, 0)
	button.EditBoxMiddle:SetTexture("Interface\\Chatframe\\UI-ChatInputBorderFocus-Mid")
	button.EditBoxMiddle:SetTexCoord(0/32, 1/32, 8/32, 25/32)
	button.EditBoxMiddle:SetVertexColor(0.6, 0.6, 0.6, 1)

	button.EditBoxRight = button.EditBox:CreateTexture(nil, "BACKGROUND")
	button.EditBoxRight:SetWidth(5)
	button.EditBoxRight:SetHeight(17)
	button.EditBoxRight:SetPoint("LEFT", button.EditBoxMiddle, "RIGHT", 0, 0)
	button.EditBoxRight:SetTexture("Interface\\Chatframe\\UI-ChatInputBorderFocus-Right")
	button.EditBoxRight:SetTexCoord(19/32, 24/32, 8/32, 25/32)
	button.EditBoxRight:SetVertexColor(0.6, 0.6, 0.6, 1)

	button.EditBoxClear = CreateFrame("Button", nil, button)
	button.EditBoxClear:SetWidth(17)
	button.EditBoxClear:SetHeight(17)
	button.EditBoxClear:SetPoint("LEFT", button.EditBox, "RIGHT", 2, 0)

	button.EditBoxClearNormal = button.EditBoxClear:CreateTexture(nil, "ARTWORK")
	button.EditBoxClearNormal:SetPoint("TOPLEFT", 0, 0)
	button.EditBoxClearNormal:SetPoint("BOTTOMRIGHT", 0, 0)
	button.EditBoxClearNormal:SetAlpha(0.75)
	button.EditBoxClearNormal:SetTexture("Interface\\FriendsFrame\\ClearBroadcastIcon")
	button.EditBoxClear:SetNormalTexture(button.EditBoxClearNormal)

	button.EditBoxClearDisable = button.EditBoxClear:CreateTexture(nil, "ARTWORK")
	button.EditBoxClearDisable:SetPoint("TOPLEFT", 0, 0)
	button.EditBoxClearDisable:SetPoint("BOTTOMRIGHT", 0, 0)
	button.EditBoxClearDisable:SetAlpha(0.25)
	button.EditBoxClearDisable:SetTexture("Interface\\FriendsFrame\\ClearBroadcastIcon")
	button.EditBoxClear:SetDisabledTexture(button.EditBoxClearDisable)

	Desaturation(button.EditBoxClearDisable, true)

	button.EditBox:SetScript("OnShow",          function(self) self:SetAutoFocus(false) end)
	button.EditBox:SetScript("OnEnterPressed",  function(self) self:ClearFocus() end)
	button.EditBox:SetScript("OnEscapePressed", function(self) self:ClearFocus() end)
	button.EditBox:SetScript("OnTextChanged",   function() funcSearch() end)

	button.EditBoxClear:Disable()
	button.EditBoxClear:SetScript("OnClick", function() funcClear() end)
	button.EditBoxClear:SetScript("OnMouseDown", function()
		button.EditBoxClearNormal:SetPoint("TOPLEFT", 1, -1)
		button.EditBoxClearNormal:SetPoint("BOTTOMRIGHT", -1, 1)
	end)
	button.EditBoxClear:SetScript("OnMouseUp", function()
		button.EditBoxClearNormal:SetPoint("TOPLEFT", 0, 0)
		button.EditBoxClearNormal:SetPoint("BOTTOMRIGHT", 0, 0)
	end)
	button.EditBoxClear:SetScript("OnEnter", function()
		button.EditBoxClearNormal:SetAlpha(1)
	end)
	button.EditBoxClear:SetScript("OnLeave", function()
		button.EditBoxClearNormal:SetAlpha(0.75)
	end)
end
-- Template EditBox END ----------------------------------------

-- Template ScrollFrame START ----------------------------------------
TEMPLATE.ButtonTextureScrollFrame = function(button, pos)
	local textureNormal = button:CreateTexture(nil, "ARTWORK")
	textureNormal:SetWidth(16)
	textureNormal:SetHeight(16)
	textureNormal:SetPoint("LEFT", 0, 0)
	textureNormal:SetTexture("Interface\\Buttons\\UI-ScrollBar-Scroll"..pos.."Button-Up")
	textureNormal:SetTexCoord(6/32, 25/32, 7/32, 24/32)
	button:SetNormalTexture(textureNormal)

	local texturePush = button:CreateTexture(nil, "ARTWORK")
	texturePush:SetWidth(16)
	texturePush:SetHeight(16)
	texturePush:SetPoint("TOPLEFT", 1, -1)
	texturePush:SetPoint("BOTTOMRIGHT", -1, 1)
	texturePush:SetTexture("Interface\\Buttons\\UI-ScrollBar-Scroll"..pos.."Button-Down")
	texturePush:SetTexCoord(6/32, 25/32, 7/32, 24/32)
	button:SetPushedTexture(texturePush)

	local textureHighlight = button:CreateTexture(nil, "ARTWORK")
	textureHighlight:SetWidth(16)
	textureHighlight:SetHeight(16)
	textureHighlight:SetPoint("LEFT", 0, 0)
	textureHighlight:SetTexture("Interface\\Buttons\\UI-ScrollBar-Scroll"..pos.."Button-Highlight")
	textureHighlight:SetTexCoord(6/32, 25/32, 7/32, 24/32)
	button:SetHighlightTexture(textureHighlight)

	local textureDisable = button:CreateTexture(nil, "ARTWORK")
	textureDisable:SetWidth(16)
	textureDisable:SetHeight(16)
	textureDisable:SetPoint("LEFT", 0, 0)
	textureDisable:SetTexture("Interface\\Buttons\\UI-ScrollBar-Scroll"..pos.."Button-Disabled")
	textureDisable:SetTexCoord(6/32, 25/32, 7/32, 24/32)
	button:SetDisabledTexture(textureDisable)
end

TEMPLATE.ScrollFrame = function(frame, child)
	frame.offset = 0
	frame:EnableMouseWheel(true)

	frame.ScrollBar = CreateFrame("Slider", nil, frame)
	frame.ScrollBar:SetWidth(16)
	frame.ScrollBar:SetHeight(0)
	frame.ScrollBar:SetPoint("TOPLEFT", frame, "TOPRIGHT", 2, -16)
	frame.ScrollBar:SetPoint("BOTTOMLEFT", frame, "BOTTOMRIGHT", 2, 16)
	frame.ScrollBar:SetMinMaxValues(0, 0)
	frame.ScrollBar:SetValue(0)
	frame.ScrollBar:SetOrientation("VERTICAL")
	frame.ScrollBar:EnableMouseWheel(true)
	frame.ScrollBar:SetScript("OnValueChanged", function(self, value)
		frame:SetVerticalScroll(value)
	end)
	frame.ScrollBar:SetScript("OnMouseWheel", function(self, delta)
		if delta > 0 then
			frame.ScrollBar:SetValue(frame.ScrollBar:GetValue() - (frame.ScrollBar:GetHeight() / 2))
		else
			frame.ScrollBar:SetValue(frame.ScrollBar:GetValue() + (frame.ScrollBar:GetHeight() / 2))
		end
	end)

	frame.ScrollBG = frame.ScrollBar:CreateTexture(nil, "BACKGROUND")
	frame.ScrollBG:SetAllPoints(frame.ScrollBar)
	frame.ScrollBG:SetTexture(0, 0, 0, 0)

	frame.ScrollBarScrollUpButton = CreateFrame("Button", nil, frame.ScrollBar)
	frame.ScrollBarScrollUpButton:SetWidth(16)
	frame.ScrollBarScrollUpButton:SetHeight(16)
	frame.ScrollBarScrollUpButton:SetPoint("BOTTOM", frame.ScrollBar, "TOP", 0, 0)
	frame.ScrollBarScrollUpButton:SetScript("OnClick", function()
		frame.ScrollBar:SetValue(frame.ScrollBar:GetValue() - (frame.ScrollBar:GetHeight() / 2))
		PlaySound("UChatScrollButton")
	end)
	TEMPLATE.ButtonTextureScrollFrame(frame.ScrollBarScrollUpButton, "Up")

	frame.ScrollBarScrollDownButton = CreateFrame("Button", nil, frame.ScrollBar)
	frame.ScrollBarScrollDownButton:SetWidth(16)
	frame.ScrollBarScrollDownButton:SetHeight(16)
	frame.ScrollBarScrollDownButton:SetPoint("TOP", frame.ScrollBar, "BOTTOM", 0, 0)
	frame.ScrollBarScrollDownButton:SetScript("OnClick", function()
		frame.ScrollBar:SetValue(frame.ScrollBar:GetValue() + (frame.ScrollBar:GetHeight() / 2))
		PlaySound("UChatScrollButton")
	end)
	TEMPLATE.ButtonTextureScrollFrame(frame.ScrollBarScrollDownButton, "Down")

	frame.ScrollBarThumbTexture = frame.ScrollBar:CreateTexture(nil, "BORDER")
	frame.ScrollBarThumbTexture:SetWidth(14)
	frame.ScrollBarThumbTexture:SetHeight(16)
	frame.ScrollBarThumbTexture:SetTexture("Interface\\Buttons\\UI-ScrollBar-Knob")
	frame.ScrollBarThumbTexture:SetTexCoord(8/32, 23/32, 8/32, 23/32)
	frame.ScrollBar:SetThumbTexture(frame.ScrollBarThumbTexture)

	frame:SetScript("OnScrollRangeChanged", function(self, xrange, yrange)
		if not yrange then
			yrange = self:GetVerticalScrollRange()
		end
		local value = frame.ScrollBar:GetValue()
		if value > yrange then
			value = yrange
		end
		frame.ScrollBar:SetMinMaxValues(0, yrange)
		frame.ScrollBar:SetValue(value)
		if math_floor(yrange) == 0 then
			frame.ScrollBarScrollDownButton:Disable()
			frame.ScrollBarScrollUpButton:Disable()
			frame.ScrollBarScrollDownButton:Show()
			frame.ScrollBarScrollUpButton:Show()
		else
			frame.ScrollBarScrollDownButton:Show()
			frame.ScrollBarScrollUpButton:Show()
			frame.ScrollBar:Show()
			-- The 0.005 is to account for precision errors
			if yrange - value > 0.005 then
				frame.ScrollBarScrollDownButton:Enable()
			else
				frame.ScrollBarScrollDownButton:Disable()
			end
		end
	end)
	frame:SetScript("OnMouseWheel", function(self, delta)
		if delta > 0 then
			frame.ScrollBar:SetValue(frame.ScrollBar:GetValue() - (frame.ScrollBar:GetHeight() / 2))
		else
			frame.ScrollBar:SetValue(frame.ScrollBar:GetValue() + (frame.ScrollBar:GetHeight() / 2))
		end
	end)
	frame:SetScript("OnVerticalScroll", function(self, off)
		frame.ScrollBar:SetValue(off)
	end)

	if child then
		frame.ScrollChildFrame = CreateFrame("Frame", nil, frame)
		frame.ScrollChildFrame:SetPoint("TOPLEFT", frame, "TOPLEFT", 0, 0)
		frame.ScrollChildFrame:SetPoint("BOTTOMLEFT", frame, "BOTTOMLEFT", 0, 0)
		frame:SetScrollChild(frame.ScrollChildFrame)
	end
end
-- Template ScrollFrame END ----------------------------------------

-- Template SortPosition START ----------------------------------------
TEMPLATE.SortPosition = function(button, text, pos, width, dummy)
	button.Background = button:CreateTexture(nil, "BACKGROUND")
	button.Background:SetPoint("TOPLEFT", 0, 0)
	button.Background:SetPoint("BOTTOMRIGHT", 0, 0)
	button.Background:SetTexture(1, 1, 1, 0.1)

	local buttontext
	if text then
		if pos == "right" then
			buttontext = button:CreateFontString(nil, "OVERLAY", "GameFontNormalSmall")
			buttontext:SetText(text)
			local fw = buttontext:GetStringWidth()
			if fw >= width then
				buttontext:SetWidth( width-10 )
			end
			buttontext:SetHeight(12)
			buttontext:SetPoint("RIGHT", button, "RIGHT", 0, 0)
			buttontext:SetJustifyH("RIGHT")
		elseif pos == "left" then
			buttontext = button:CreateFontString(nil, "OVERLAY", "GameFontNormalSmall")
			buttontext:SetText(text)
			local fw = buttontext:GetStringWidth()
			if fw >= width then
				buttontext:SetWidth( width-10 )
			end
			buttontext:SetHeight(12)
			buttontext:SetPoint("LEFT", button, "LEFT", 0, 0)
			buttontext:SetJustifyH("LEFT")
		elseif pos == "empty" then
			buttontext = button:CreateFontString(nil, "OVERLAY", "GameFontNormalSmall")
			buttontext:SetText(text)
			buttontext:SetWidth(0.01)
			buttontext:SetHeight(12)
			buttontext:SetPoint("CENTER", button, "CENTER", 0, 0)
			buttontext:SetJustifyH("CENTER")
		end
	end

	if dummy then return end

	button.arrow = button:CreateTexture(nil, "BACKGROUND")
	button.arrow:SetWidth(7)
	button.arrow:SetHeight(8)
	button.arrow:SetTexture("Interface\\Buttons\\UI-SortArrow")
	if text then
		if pos == "right" then
			button.arrow:SetPoint("RIGHT", buttontext, "LEFT", -1, 0)
		elseif pos == "left" then
			button.arrow:SetPoint("LEFT", buttontext, "RIGHT", 1, 0)
		elseif pos == "empty" then
			button.arrow:SetPoint("CENTER", button, "CENTER", 0, 0)
		end
	else
		button.arrow:SetPoint("CENTER", button, "CENTER", 0, 0)
	end

	button.Highlight = button:CreateTexture(nil, "ARTWORK")
	button.Highlight:SetPoint("TOPLEFT", 0, 0)
	button.Highlight:SetPoint("BOTTOMRIGHT", 0, 0)
	button.Highlight:SetTexture(1, 1, 1, 0.2)
	button:SetHighlightTexture(button.Highlight)
end
-- Template SortPosition END ----------------------------------------

-- Template DataList START ----------------------------------------
TEMPLATE.DataList = function(button, isSort)
	if isSort then
		button.NumberSort = CreateFrame("Button", nil, button)
		TEMPLATE.SortPosition(button.NumberSort, nil, "empty", width_DataList_Number, "dummy")
		button.NumberSort:SetWidth(width_DataList_Number)
		button.NumberSort:SetHeight(12)
		button.NumberSort:SetPoint("LEFT", button, "LEFT", 0, 0)

		button.NPCSort = CreateFrame("Button", nil, button)
		TEMPLATE.SortPosition(button.NPCSort, L["NPC-ID"], "right", width_DataList_NPC)
		button.NPCSort:SetWidth(width_DataList_NPC)
		button.NPCSort:SetHeight(12)
		button.NPCSort:SetPoint("LEFT", button, "LEFT", width_DataList_Number+2, 0)
		button.NPCSort:SetScript("OnClick", function() FlightMaster:GUIsort(1) end)

		button.ExceptionsSort = CreateFrame("Button", nil, button)
		TEMPLATE.SortPosition(button.ExceptionsSort, "", "empty", width_DataList_Exceptions)
		button.ExceptionsSort:SetWidth(width_DataList_Exceptions)
		button.ExceptionsSort:SetHeight(12)
		button.ExceptionsSort:SetPoint("LEFT", button, "LEFT", width_DataList_Number+width_DataList_NPC+2+2, 0)
		button.ExceptionsSort:SetScript("OnClick", function() FlightMaster:GUIsort(2) end)

		button.FlightPointSort = CreateFrame("Button", nil, button)
		TEMPLATE.SortPosition(button.FlightPointSort, L["Flight point"], "left", 220)
		button.FlightPointSort:SetWidth(420-width_DataList_Number-width_DataList_NPC-width_DataList_Exceptions -offset-2-2-2-16-offset)
		button.FlightPointSort:SetHeight(12)
		button.FlightPointSort:SetPoint("LEFT", button, "LEFT", width_DataList_Number+width_DataList_NPC+width_DataList_Exceptions+2+2+2, 0)
		button.FlightPointSort:SetScript("OnClick", function() FlightMaster:GUIsort(3) end)

	else
		button.NumberButton = button:CreateFontString(nil, "ARTWORK", "GameFontNormalSmall")
		button.NumberButton:SetWidth(width_DataList_Number)
		button.NumberButton:SetHeight(12)
		button.NumberButton:SetPoint("LEFT", button, "LEFT", 0, 0)
		button.NumberButton:SetJustifyH("RIGHT")
		button.NumberButton:SetTextColor(1, 1, 0.49, 1)

		button.NPCButton = button:CreateFontString(nil, "ARTWORK", "GameFontNormalSmall")
		button.NPCButton:SetWidth(width_DataList_NPC)
		button.NPCButton:SetHeight(12)
		button.NPCButton:SetPoint("LEFT", button, "LEFT", width_DataList_Number+2, 0)
		button.NPCButton:SetJustifyH("RIGHT")
		button.NPCButton:SetTextColor(1, 1, 1, 1)

		button.ExceptionsButtonButton = CreateFrame("Button", nil, button)
		button.ExceptionsButtonButton:SetWidth(width_DataList_Exceptions)
		button.ExceptionsButtonButton:SetHeight(12)
		button.ExceptionsButtonButton:SetPoint("LEFT", button, "LEFT", width_DataList_Number+width_DataList_NPC+2+2, 0)
		button.ExceptionsButtonButton:SetFrameLevel( button:GetFrameLevel()+2 )
		button.ExceptionsButtonButton:SetScript("OnEnter", function()
			button:LockHighlight()
			FlightMaster:NodeHighlight(button.node, 1)
			FlightMaster:ShowException(button, 1)
		end)
		button.ExceptionsButtonButton:SetScript("OnLeave", function()
			button:UnlockHighlight()
			FlightMaster:NodeHighlight()
			FlightMaster:ShowException(button)
		end)

		button.ExceptionsButton = button.ExceptionsButtonButton:CreateFontString(nil, "ARTWORK", "GameFontNormalSmall")
		button.ExceptionsButton:SetWidth(width_DataList_Exceptions)
		button.ExceptionsButton:SetHeight(12)
		button.ExceptionsButton:SetPoint("LEFT", button, "LEFT", width_DataList_Number+width_DataList_NPC+2+2, 0)
		button.ExceptionsButton:SetJustifyH("CENTER")
		button.ExceptionsButton:SetTextColor(1, 1, 0.49, 1)

		button.FlightPointButton = button:CreateFontString(nil, "ARTWORK", "GameFontNormalSmall")
		button.FlightPointButton:SetWidth(420-width_DataList_Number-width_DataList_NPC-width_DataList_Exceptions -offset-2-2-2-16-offset)
		button.FlightPointButton:SetHeight(12)
		button.FlightPointButton:SetPoint("LEFT", button, "LEFT", width_DataList_Number+width_DataList_NPC+width_DataList_Exceptions+2+2+2, 0)
		button.FlightPointButton:SetJustifyH("LEFT")

		button.Highlight = button:CreateTexture(nil, "ARTWORK")
		button.Highlight:SetPoint("TOPLEFT", button, "TOPLEFT", 0, 0)
		button.Highlight:SetPoint("BOTTOMRIGHT", button, "BOTTOMRIGHT", 0, 0)---26, 0)
		button.Highlight:SetTexture(1, 1, 1, 0.2)
		button:SetHighlightTexture(button.Highlight)

		button:SetScript("OnEnter", function() FlightMaster:NodeHighlight(button.node, 1) end)
		button:SetScript("OnLeave", function() FlightMaster:NodeHighlight()               end)
	end
end
-- Template DataList END ----------------------------------------
-- ---------------------------------------------------------------------------------------------------------------------

-- ---------------------------------------------------------------------------------------------------------------------
function FlightMaster:InitOptions()
	SlashCmdList["FLIGHTMASTER"] = function() FlightMaster:MainFrameToggle() end
	SLASH_FLIGHTMASTER1 = "/flightmaster"
	SLASH_FLIGHTMASTER2 = "/fm"

	FlightMaster_Options.UTF8force = L["UTF8force"]

	-- FlightMaster version check
	if FlightMaster_PlayerNodes then
		if FlightMaster_PlayerNodes.version == nil then
			FlightMaster_PlayerNodes.version = 2
			FlightMaster_PlayerNodes.UTF8force = nil
			local newNode = 1597354 -- Pandaria: Halfhill, Valley of the Four Winds
			local oldNode = 1589353
			for map in pairs(FlightMaster_PlayerNodes) do
				if map == "Pandaria" then
					for node in pairs(FlightMaster_PlayerNodes[map]) do
						if node == oldNode then
							local oldData = FlightMaster_PlayerNodes[map][oldNode]
							--print("playerNodes:", oldData)
							FlightMaster_PlayerNodes[map][oldNode] = nil
							FlightMaster_PlayerNodes[map][newNode] = oldData
							break
						end
					end
				end
			end
		end
	end
	if FlightMaster_Options.version == nil then
		FlightMaster_Options.version = 2
	end
	if FlightMaster_Options.version == 1 then
		local newNode = 1597354 -- Pandaria: Halfhill, Valley of the Four Winds
		local oldNode = 1589353
		if FlightMaster_Nodes then
			for loc in pairs(FlightMaster_Nodes) do
				for faction in pairs(FlightMaster_Nodes[loc]) do
					for map in pairs(FlightMaster_Nodes[loc][faction]) do
						if map == "Pandaria" then
							for coord, name in pairs(FlightMaster_Nodes[loc][faction][map]) do
								if coord == oldNode then
									local oldName = FlightMaster_Nodes[loc][faction][map][oldNode]
									--print("oldName:", oldName)
									FlightMaster_Nodes[loc][faction][map][oldNode] = nil
									FlightMaster_Nodes[loc][faction][map][newNode] = oldName
									break
								end
							end
						end
					end
				end
			end
		end
		if FlightMaster_Data then
			-- destination
			for faction in pairs(FlightMaster_Data) do
				for map in pairs(FlightMaster_Data[faction]) do
					if map == "Pandaria" then
						for startNode in pairs(FlightMaster_Data[faction][map]) do
							for destination, data in pairs(FlightMaster_Data[faction][map][startNode]) do
								if destination == oldNode then
									--print("data", startNode, oldNode, data)
									FlightMaster_Data[faction][map][startNode][newNode] = data
									FlightMaster_Data[faction][map][startNode][oldNode] = nil
								end
							end
						end
					end
				end
			end
			-- start
			for faction in pairs(FlightMaster_Data) do
				for map in pairs(FlightMaster_Data[faction]) do
					if map == "Pandaria" then
						for startNode in pairs(FlightMaster_Data[faction][map]) do
							if startNode == oldNode then
								local oldData = FlightMaster_Data[faction][map][oldNode]
								--print("oldData:", oldNode, oldData)
								FlightMaster_Data[faction][map][oldNode] = nil
								FlightMaster_Data[faction][map][newNode] = oldData
								break
							end
						end
					end
				end
			end
		end
		FlightMaster_Options.version = 2
	end
	-- FlightMaster version check

	-- Pandaria remapping 50400 - Patch: 5.4.0.17205
	if not FlightMaster_Options.v50400remapped then
		local NEW = {
			-- old        new
			[1478336] = 1411336, -- NEUTRAL  | Stoneplow, Valley of the Four Winds
			[1791499] = 1723499, -- NEUTRAL  | Dawn's Blossom, Jade Forest
			[1835438] = 1768438, -- NEUTRAL  | Jade Temple Grounds, Jade Forest
			[1849507] = 1782507, -- NEUTRAL  | The Arboretum, Jade Forest
			[1840587] = 1773587, -- NEUTRAL  | Sri-La Village, Jade Forest
			[1813574] = 1746574, -- NEUTRAL  | Emperor's Omen, Jade Forest
			[1770583] = 1703583, -- NEUTRAL  | Tian Monastery, Jade Forest
			[1690418] = 1623418, -- NEUTRAL  | Pang's Stead, Valley of the Four Winds
			[1597354] = 1530354, -- NEUTRAL  | Halfhill, Valley of the Four Winds
			[1691341] = 1624341, -- NEUTRAL  | Zhu's Watch, Krasarang Wilds
			[1455601] = 1387601, -- NEUTRAL  | Kota Basecamp, Kun-Lai Summit
			[1595162] = 1528162, -- NEUTRAL  | Marista, Krasarang Wilds
			[1578667] = 1511667, -- NEUTRAL  | Temple of the White Tiger, Kun-Lai Summit
			[1461531] = 1394531, -- NEUTRAL  | Shado-Pan Fallback, Kun-Lai Summit
			[1611515] = 1544515, -- NEUTRAL  | Binan Village, Kun-Lai Summit
			[1534635] = 1466635, -- NEUTRAL  | One Keg, Kun-Lai Summit
			[1558739] = 1491739, -- NEUTRAL  | Zouchin Village, Kun-Lai Summit
			[1411638] = 1344638, -- NEUTRAL  | Winter's Blossom, Kun-Lai Summit
			[1654439] = 1586439, -- NEUTRAL  | Tavern in the Mists, The Veiled Stair
			[1645411] = 1578411, -- NEUTRAL  | Grassy Cline, Valley of the Four Winds
			[1303501] = 1236501, -- NEUTRAL  | Rensai's Watchpost, Townlong Steppes
			[1385571] = 1317571, -- NEUTRAL  | Longying Outpost, Townlong Steppes
			[1401494] = 1333494, -- NEUTRAL  | Gao-Ran Battlefront, Townlong Steppes
			[1283524] = 1216524, -- NEUTRAL  | Shado-Pan Garrison, Townlong Steppes
			[1458424] = 1391424, -- NEUTRAL  | Serpent's Spine, Vale of Eternal Blossoms
			[1373281] = 1306281, -- NEUTRAL  | Soggy's Gamble, Dread Wastes
			[1346455] = 1279455, -- NEUTRAL  | The Sunset Brewgarden, Dread Wastes
			[1371387] = 1304387, -- NEUTRAL  | Klaxxi'vess, Dread Wastes
			[1768412] = 1700412, -- NEUTRAL  | Serpent's Overlook, Jade Forest
			[1512198] = 1445198, -- NEUTRAL  | Cradle of Chi-Ji, Krasarang Wilds
			[1312325] = 1245325, -- NEUTRAL  | The Briny Muck, Dread Wastes
			[1622299] = 1555299, -- HORDE    | Thunder Cleft, Krasarang Wilds
			[1503231] = 1436231, -- HORDE    | Dawnchaser Retreat, Krasarang Wilds
			[1678492] = 1610492, -- HORDE    | Grookin Hill, Jade Forest
			[1561505] = 1494505, -- HORDE    | Shrine of Two Moons, Vale of Eternal Blossoms
			[1558562] = 1491562, -- HORDE    | Eastwind Rest, Kun-Lai Summit
			[1419551] = 1352551, -- HORDE    | Serpent's Spine, Kun-Lai Summit
			[1679619] = 1612619, -- HORDE    | Honeydew Village, Jade Forest
			[1427225] = 1360225, -- HORDE    | Domination Point, Krasarang Wilds
			[1546860] = 1478860, -- HORDE    | Bozzle's Wreck, Isle Of Giants
			[1855357] = 1787357, -- ALLIANCE | Pearlfin Village, Jade Forest
			[1488275] = 1420275, -- ALLIANCE | Sentinel Basecamp, Krasarang Wilds
			[1656278] = 1588278, -- ALLIANCE | The Incursion, Krasarang Wilds
			[1608448] = 1541448, -- ALLIANCE | Shrine of Seven Stars, Vale of Eternal Blossoms
			[1513549] = 1446549, -- ALLIANCE | Westwind Rest, Kun-Lai Summit
			[1452392] = 1384392, -- ALLIANCE | The Lion's Redoubt, Dread Wastes
			[1785347] = 1718347, -- ALLIANCE | Paw'Don Village, Jade Forest
			[1737272] = 1670272, -- ALLIANCE | Lion's Landing, Krasarang Wilds
			[1531856] = 1464856, -- ALLIANCE | Beeble's Wreck, Isle Of Giants
		}

		if FlightMaster_PlayerNodes and FlightMaster_PlayerNodes["Pandaria"] then
			for oldNode, newNode in pairs(NEW) do
				for node in pairs(FlightMaster_PlayerNodes["Pandaria"]) do
					if node == oldNode then
						local oldData = FlightMaster_PlayerNodes["Pandaria"][oldNode]
						--print("playerNodes:", oldData)
						FlightMaster_PlayerNodes["Pandaria"][oldNode] = nil
						FlightMaster_PlayerNodes["Pandaria"][newNode] = oldData
						break
					end
				end
			end
		end

		if FlightMaster_Nodes then
			for loc in pairs(FlightMaster_Nodes) do
				for faction in pairs(FlightMaster_Nodes[loc]) do
					for map in pairs(FlightMaster_Nodes[loc][faction]) do
						if map == "Pandaria" then
							for coord, name in pairs(FlightMaster_Nodes[loc][faction][map]) do
								for oldNode, newNode in pairs(NEW) do
									if coord == oldNode then
										local oldName = FlightMaster_Nodes[loc][faction][map][oldNode]
										--print("Nodes:", oldNode, newNode, oldName)
										FlightMaster_Nodes[loc][faction][map][oldNode] = nil
										FlightMaster_Nodes[loc][faction][map][newNode] = oldName
										break
									end
								end
							end
						end
					end
				end
			end
		end
		if FlightMaster_Data then
			for oldNode, newNode in pairs(NEW) do
				-- destination
				for faction in pairs(FlightMaster_Data) do
					for map in pairs(FlightMaster_Data[faction]) do
						if map == "Pandaria" then
							for startNode in pairs(FlightMaster_Data[faction][map]) do
								for destination, data in pairs(FlightMaster_Data[faction][map][startNode]) do
									if destination == oldNode then
										--print("Data (dest)", oldNode, newNode, startNode, data)
										FlightMaster_Data[faction][map][startNode][newNode] = data
										FlightMaster_Data[faction][map][startNode][oldNode] = nil
									end
								end
							end
						end
					end
				end
				-- start
				for faction in pairs(FlightMaster_Data) do
					for map in pairs(FlightMaster_Data[faction]) do
						if map == "Pandaria" then
							for startNode in pairs(FlightMaster_Data[faction][map]) do
								if startNode == oldNode then
									local oldData = FlightMaster_Data[faction][map][oldNode]
									--print("Data: (start)", oldNode, newNode, oldData)
									FlightMaster_Data[faction][map][oldNode] = nil
									FlightMaster_Data[faction][map][newNode] = oldData
									break
								end
							end
						end
					end
				end
			end
		end

		FlightMaster_Options.v50400remapped = true
	end

	if FlightMaster_Options.DataVersion == nil then FlightMaster_Options.DataVersion = 0 end



	-- -------------------------------------------------------------------------------------------------------------------
	-- ------------------------------------------------------- --
	-- THIS IS FOR MAINTENANCE ONLY - DO NOT UNCOMMENT - START --
	-- ------------------------------------------------------- --

	--[[ -- make DEFAULT data START --------------------------------------------------------
	print("1 ---------- START make DEFAULT data START")
	local table_insert = table.insert
	local table_sort = table.sort
	local function SortedPairs(t)
		local ntab = {}
		for k,v in pairs(t) do table_insert(ntab,k) end
		table_sort(ntab, function(a,b) if a < b then return true end end)
		local i = 0
		local function _f(_s,_v)
			i = i + 1
			local k = ntab[i]
			if k then return k,t[k] end
		end
		return _f,nil,nil
	end
	local num = 0
	local lines = 0
	local str= '-- -------------------------------------------------------------------------- --\n' lines = lines + 1
	str = str..'-- FlightMaster (DEFAULT DATA)                                                --\n' lines = lines + 1
	str = str..'-- -------------------------------------------------------------------------- --\n' lines = lines + 1
	str = str..'\n' lines = lines + 1
	str = str..'FlightMaster_DEFAULT = {\n' lines = lines + 1
	str = str..'\t["VERSION"] = '..(FlightMaster_Options.DataVersion + 1)..',\n' lines = lines + 1
	str = str..'\t["DEFAULT"] = {\n' lines = lines + 1
	for faction, v in SortedPairs(FlightMaster_Data) do
		str = str..'\t\t["'..faction..'"] = {\n' lines = lines + 1
		for map, v2 in SortedPairs(v) do
			str = str..'\t\t\t["'..map..'"] = {\n' lines = lines + 1
			for start, v3 in SortedPairs(v2) do
				str = str..'\t\t\t\t['..start..'] = {\n' lines = lines + 1
				for destination, value in SortedPairs(v3) do
					local time, taken = string_match(value, ":(%d+):(%d+):")
					time = tonumber(time)
					if time > 0 then
						str = str..'\t\t\t\t\t['..destination..'] = '..time..',\n' lines = lines + 1
						num = num + 1
					end
				end
				str = str..'\t\t\t\t},\n' lines = lines + 1
			end
			str = str..'\t\t\t},\n' lines = lines + 1
		end
		str = str..'\t\t},\n' lines = lines + 1
	end
	str = str..'\t},\n' lines = lines + 1
	str = str..'}' lines = lines + 1
	GVAR.DefaultFrame = CreateFrame("Frame", nil, UIParent)
	TEMPLATE.BorderTRBL(GVAR.DefaultFrame)
	GVAR.DefaultFrame:EnableMouse(true)
	GVAR.DefaultFrame:SetToplevel(true)
	GVAR.DefaultFrame:SetPoint("CENTER")
	GVAR.DefaultFrame:SetWidth(500)
	GVAR.DefaultFrame:SetHeight(80)
	GVAR.DefaultFrame:Show()
	GVAR.DefaultFrameEditbox = CreateFrame("EditBox", nil, GVAR.DefaultFrame, "InputBoxTemplate")
	GVAR.DefaultFrameEditbox:Insert(str)
	GVAR.DefaultFrameEditbox:SetWidth(425)
	GVAR.DefaultFrameEditbox:SetHeight(16)
	GVAR.DefaultFrameEditbox:SetPoint("BOTTOM", GVAR.DefaultFrame, "BOTTOM", 0, 10)
	GVAR.DefaultFrameEditbox:SetMultiLine(false)
	GVAR.DefaultFrameEditbox:SetFontObject(NumberFontNormalSmall)
	GVAR.DefaultFrameEditbox:ClearFocus()
	GVAR.DefaultFrameEditbox:SetAutoFocus(true)
	GVAR.DefaultFrameEditbox:HighlightText()
	GVAR.DefaultFrameEditbox:SetCursorPosition(0)
	GVAR.DefaultFrameTxt = GVAR.DefaultFrame:CreateFontString(nil, "ARTWORK", "GameFontNormal")
	GVAR.DefaultFrameTxt:SetPoint("BOTTOM", GVAR.DefaultFrameEditbox, "TOP", 0, 10)
	GVAR.DefaultFrameTxt:SetJustifyH("CENTER")
	GVAR.DefaultFrameTxt:SetText("make DEFAULT\neverything is selected -> copy & paste\nflight duration data: "..num.." - lines: "..lines)
	print("make DEFAULT - flight duration data:", num)
	print("make DEFAULT - lines:", lines)
	print("1 ---------- END make DEFAULT data END")
	--]] -- make DEFAULT data END --------------------------------------------------------

	--[[ ---------------------------------------------------------------------------------
	print("2 ---------- START NPC count START")
	for faction, value in pairs(FlightMaster_NPC) do
		local total = 0
		for map, value2 in pairs(value) do
			local count = 0
			for npc, startCoord in pairs(value2) do
				count = count + 1
				total = total + 1
			end
			print("NPC count:", faction, map, "=", count)
		end
		print("NPC count:", faction, "--- Total =", total)
	end
	print("2 ---------- END NPC count END")

	print("3 ---------- START check if start/destination from DEFAULT is not in NPC data START")
	local sd = {} -- build one table that contains all start and destination points from DEFAULT
	for faction, value in pairs(FlightMaster_DEFAULT["DEFAULT"]) do
		for map, value2 in pairs(value) do
			for start, value3 in pairs(value2) do
				sd[start] = 1 --print(start)
				for destination in pairs(value3) do
					sd[destination] = 1 --print(destination)
				end
			end
		end
	end
	local sdx = {} -- build table with all start points from the master NPC table
	for faction, value in pairs(FlightMaster_NPC) do
		for map, value2 in pairs(value) do
			for npc, start in pairs(value2) do
				sdx[start] = 1 --print(start)
			end
		end
	end
	for k, v in pairs(sd) do -- iterate over all start and destination points and check if it is a valid value from master
		if not sdx[k] then
			print("unknown start/dest:", k)
		end
	end
	print("3 ---------- END check if start/destination from DEFAULT is not in NPC data END")

	print("4 ---------- START isInDEFAULT check START")
	local sd = {}
	local sdCount = 0
	for faction, value in pairs(FlightMaster_DEFAULT["DEFAULT"]) do
		for map, value2 in pairs(value) do
			local aNum = 0
			local check = {}
			for start, value3 in pairs(value2) do
				if not sd[start] then
					sd[start] = 1
					sdCount = sdCount + 1
				end
				for destination in pairs(value3) do
					if not sd[destination] then
						sd[destination] = 1
						sdCount = sdCount + 1
					end
				end
				for eNPC, eStart in pairs(FlightMaster_NPC[faction][map]) do
					if start == eStart then
						aNum = aNum + 1
						check[eNPC] = 1
					end
				end
			end
			print("isInDEFAULT check:", faction, map, "=", aNum)
			for eNPC, eStart in pairs(FlightMaster_NPC[faction][map]) do
				if not check[eNPC] then
					local name = "?"
					if FlightMaster_Nodes[locale] and
					   FlightMaster_Nodes[locale][faction] and
					   FlightMaster_Nodes[locale][faction][map] and
					   FlightMaster_Nodes[locale][faction][map][eStart]
					then
						name = FlightMaster_Nodes[locale][faction][map][eStart]
					end
					print("isInDEFAULT check: no DEFAULT start from here:", faction, "#", eNPC, "#", eStart, "# name =", name)
				end
			end
		end
	end
	print("4 ---------- END isInDEFAULT check END")

	print("5 ---------- START dura differ - checks for differences between DEFAULT and player dura START")
	for faction, value in pairs(FlightMaster_DEFAULT["DEFAULT"]) do
		for map, value2 in pairs(value) do
			for start, value3 in pairs(value2) do
				for destination, duration in pairs(value3) do
					if FlightMaster_Data[faction] and
					   FlightMaster_Data[faction][map] and
					   FlightMaster_Data[faction][map][start] and
					   FlightMaster_Data[faction][map][start][destination]
					then
						local time, taken = string_match(FlightMaster_Data[faction][map][start][destination], ":(%d+):(%d+):")
						time = tonumber(time)
						if duration ~= time then
							local dif = -(duration - time)

							if dif < -1 or dif > 1 then
								local startName = "?"
								if FlightMaster_Nodes and
								   FlightMaster_Nodes[locale] and
								   FlightMaster_Nodes[locale][faction] and
								   FlightMaster_Nodes[locale][faction][map] and
								   FlightMaster_Nodes[locale][faction][map][start]
								then
									startName = FlightMaster_Nodes[locale][faction][map][start]
								end
								local destinationName = "?"
								if FlightMaster_Nodes and
								   FlightMaster_Nodes[locale] and
								   FlightMaster_Nodes[locale][faction] and
								   FlightMaster_Nodes[locale][faction][map] and
								   FlightMaster_Nodes[locale][faction][map][destination]
								then
									destinationName = FlightMaster_Nodes[locale][faction][map][destination]
								end
								
								print(
									"dura differ", faction, map, "#",
									start, "("..startName..") ->", destination, "("..destinationName..")", "#",
									"DEFAULT:", duration, "player:", time, "# DIF:", dif
								)
							end

						end
					end
				end
			end
		end
	end
	print("5 ---------- END dura differ - checks for differences between DEFAULT and player dura END")

	print("6 ---------- START DEFAULT top 25 count START")
	local tcount = {}
	local count, Acount, Hcount = 0, 0, 0
	for faction, value in pairs(FlightMaster_DEFAULT["DEFAULT"]) do
		for map, value2 in pairs(value) do
			local Mcount = 0
			for start, value3 in pairs(value2) do
				for destination, flightTime in pairs(value3) do
					count = count + 1
					Mcount = Mcount + 1
					if faction == "Alliance" then
						Acount = Acount + 1
					elseif faction == "Horde" then
						Hcount = Hcount + 1
					end
					tcount[count] = {}
					tcount[count].faction = faction
					tcount[count].map = map
					tcount[count].start = start
					tcount[count].destination = destination
					tcount[count].flightTime = flightTime
					local startName = "?"
					if FlightMaster_Nodes and
					   FlightMaster_Nodes[locale] and
					   FlightMaster_Nodes[locale][faction] and
					   FlightMaster_Nodes[locale][faction][map] and
					   FlightMaster_Nodes[locale][faction][map][start]
					then
						startName = FlightMaster_Nodes[locale][faction][map][start]
					end
					local destinationName = "?"
					if FlightMaster_Nodes and
					   FlightMaster_Nodes[locale] and
					   FlightMaster_Nodes[locale][faction] and
					   FlightMaster_Nodes[locale][faction][map] and
					   FlightMaster_Nodes[locale][faction][map][destination]
					then
						destinationName = FlightMaster_Nodes[locale][faction][map][destination]
					end
					tcount[count].startName = startName
					tcount[count].destinationName = destinationName
				end
			end
			print("DEFAULT data count:", faction, map, Mcount)
		end
	end
	local Aper = math.ceil((Acount/count)*100)
	local Hper = 100-Aper
	print("DEFAULT data count total:", count, "Horde:", Hcount, "("..Hper.."%)", "Alliance:", Acount, "("..Aper.."%)")
	table_sort(tcount, function(a,b) if a.flightTime > b.flightTime then return true end end)
	print("DEFAULT top25: (long)")
	for i = 1, 25 do
		print(i, "#", tcount[i].flightTime,
		"#", tcount[i].faction,
		"#", tcount[i].map,
		"#", tcount[i].start, "("..tcount[i].startName..")", "->", tcount[i].destination, "("..tcount[i].destinationName..")")
	end
	table_sort(tcount, function(a,b) if a.flightTime < b.flightTime then return true end end)
	print("DEFAULT top25: (short)")
	for i = 1, 25 do
		print(i, "#", tcount[i].flightTime,
		"#", tcount[i].faction,
		"#", tcount[i].map,
		"#", tcount[i].start, "("..tcount[i].startName..")", "->", tcount[i].destination, "("..tcount[i].destinationName..")")
	end
	print("6 ---------- END DEFAULT top 25 count END")
	--]] ---------------------------------------------------------------------------------

	--if not sdisdtzsidisdb then return end

	-- ------------------------------------------------------- --
	-- THIS IS FOR MAINTENANCE ONLY - DO NOT UNCOMMENT - END   --
	-- ------------------------------------------------------- --
	-- -------------------------------------------------------------------------------------------------------------------



	-- insert new default flight data START
	if FlightMaster_DEFAULT and FlightMaster_DEFAULT["VERSION"] > FlightMaster_Options.DataVersion then
		FlightMaster_Options.DataVersion = FlightMaster_DEFAULT["VERSION"]
		local num = 0
		for faction, value in pairs(FlightMaster_DEFAULT["DEFAULT"]) do
			if not FlightMaster_Data[faction] then FlightMaster_Data[faction] = {} end
			for map, value2 in pairs(value) do
				if not FlightMaster_Data[faction][map] then FlightMaster_Data[faction][map] = {} end
				for start, value3 in pairs(value2) do
					if not FlightMaster_Data[faction][map][start] then FlightMaster_Data[faction][map][start] = {} end
					for destination, flightTime in pairs(value3) do
						-- ----------
						if FlightMaster_Data[faction][map][start][destination] then
							local existingTime, existingTaken = string_match(FlightMaster_Data[faction][map][start][destination], ":(%d+):(%d+):")
							existingTime = tonumber(existingTime)
							existingTaken = tonumber(existingTaken)
							if existingTaken == 0 and flightTime ~= existingTime then
								--print("UPD existingTaken == 0", "existingTime:", existingTime, "flightTime:", flightTime, "existingTaken:", existingTaken)
								FlightMaster_Data[faction][map][start][destination] = ":"..flightTime..":"..existingTaken..":"
								num = num + 1
							elseif existingTime <= 3 then
								--print("UPD existingTime <= 3", "existingTime:", existingTime, "flightTime:", flightTime, "existingTaken:", existingTaken)
								FlightMaster_Data[faction][map][start][destination] = ":"..flightTime..":"..existingTaken..":"
								num = num + 1
							end
						else
							FlightMaster_Data[faction][map][start][destination] = ":"..flightTime..":0:"
							num = num + 1
						end
						-- ----------
					end
				end
			end
		end
		if num > 0 then
			Print(string_format(L["Flight duration: %s new times added."], "|cffffff82"..num.."|r"))
		end
	end
	FlightMaster_DEFAULT = nil
	-- insert new default flight data END

	-- localized node names: current locale only
	for tableLocale, v in pairs(FlightMaster_Nodes) do
		if tableLocale == locale then
			FlightMaster_local_Nodes[tableLocale] = v --print("FlightMaster_local_Nodes:", tableLocale) -- TEST
		end
	end

	if FlightMaster_Options.MinimapButton    == nil then FlightMaster_Options.MinimapButton    = false end
	if FlightMaster_Options.MinimapButtonPos == nil then FlightMaster_Options.MinimapButtonPos = -90   end
	if FlightMaster_Options.MissingFP        == nil then FlightMaster_Options.MissingFP        = false end
	if FlightMaster_Options.CorrectFL        == nil then FlightMaster_Options.CorrectFL        = false end
	if FlightMaster_Options.Overview         == nil then FlightMaster_Options.Overview         = false end
	if FlightMaster_Options.WorldView        == nil then FlightMaster_Options.WorldView        = 1     end
end
-- ---------------------------------------------------------------------------------------------------------------------

-- ---------------------------------------------------------------------------------------------------------------------
function FlightMaster:LDBcheck()
	if LibStub and LibStub:GetLibrary("CallbackHandler-1.0", true) and LibStub:GetLibrary("LibDataBroker-1.1", true) then
		LibStub:GetLibrary("LibDataBroker-1.1"):NewDataObject("FlightMaster", {
			type = "launcher",
			icon = AddonIcon,
			OnClick = function(self, button)
				FlightMaster:MainFrameToggle()
			end,
		})
	end
end
-- ---------------------------------------------------------------------------------------------------------------------

-- ---------------------------------------------------------------------------------------------------------------------
function FlightMaster:CreateInterfaceOptions()
	GVAR.InterfaceOptions = CreateFrame("Frame", "FlightMaster_InterfaceOptions")
	GVAR.InterfaceOptions.name = "FlightMaster"

	GVAR.InterfaceOptions.Title = GVAR.InterfaceOptions:CreateFontString(nil, "ARTWORK", "GameFontNormalLarge")
	GVAR.InterfaceOptions.Title:SetPoint("TOPLEFT", 16, -16)
	GVAR.InterfaceOptions.Title:SetText("FlightMaster")
	GVAR.InterfaceOptions.Title:SetJustifyH("LEFT")
	GVAR.InterfaceOptions.Title:SetJustifyV("TOP")

	GVAR.InterfaceOptions.GUI = CreateFrame("Button", nil, GVAR.InterfaceOptions)
	TEMPLATE.TextButton(GVAR.InterfaceOptions.GUI, L["Open GUI"], 1)
	GVAR.InterfaceOptions.GUI:SetWidth(150)
	GVAR.InterfaceOptions.GUI:SetHeight(22)
	GVAR.InterfaceOptions.GUI:SetPoint("TOPLEFT", GVAR.InterfaceOptions.Title, "BOTTOMLEFT", 0, -10)
	GVAR.InterfaceOptions.GUI:SetScript("OnClick", function()
		InterfaceOptionsFrame_Show()
		HideUIPanel(GameMenuFrame)
		FlightMaster:MainFrameToggle()
	end)

	GVAR.InterfaceOptions.SlashCommandText = GVAR.InterfaceOptions:CreateFontString(nil, "ARTWORK", "GameFontNormalSmall")
	GVAR.InterfaceOptions.SlashCommandText:SetPoint("LEFT", GVAR.InterfaceOptions.GUI, "RIGHT", 10, 0)
	GVAR.InterfaceOptions.SlashCommandText:SetText("/flightmaster - /fm")
	GVAR.InterfaceOptions.SlashCommandText:SetTextColor(1, 1, 0.49, 1)
	GVAR.InterfaceOptions.SlashCommandText:SetNonSpaceWrap(true)

	GVAR.InterfaceOptions.MinimapButton = CreateFrame("CheckButton", nil, GVAR.InterfaceOptions)
	TEMPLATE.CheckButton(GVAR.InterfaceOptions.MinimapButton, 16, 4, L["Minimap-Button"])
	GVAR.InterfaceOptions.MinimapButton:SetPoint("TOPLEFT", GVAR.InterfaceOptions.GUI, "BOTTOMLEFT", 0, -10)
	GVAR.InterfaceOptions.MinimapButton:SetChecked(FlightMaster_Options.MinimapButton)
	GVAR.InterfaceOptions.MinimapButton:SetScript("OnClick", function()
		FlightMaster_Options.MinimapButton = not FlightMaster_Options.MinimapButton
		FlightMaster:CreateMinimapButton()
	end)

	InterfaceOptions_AddCategory(GVAR.InterfaceOptions)
end
-- ---------------------------------------------------------------------------------------------------------------------

-- ---------------------------------------------------------------------------------------------------------------------
local BarUpdate_TimerStart
local BarUpdate_TimerDuration
local throttle = 0.33
local elapsed = 0

local function BarUpdate(self, elap)
	elapsed = elapsed + elap
	if elapsed < throttle then return end
	elapsed = 0
	local now = GetTime()
	local endTime = BarUpdate_TimerStart + BarUpdate_TimerDuration
	local duration = endTime - now
	if now >= endTime then
		local overrun = now - endTime
		GVAR.BarFrame.DurationText:SetText(L["Time"]..": |cffff0000"..string_format("%d:%02d", overrun/60, math_fmod(overrun, 60))..flightEstimate.."|r"..flightHops.." - "..flightCost)
	else
		GVAR.BarFrame.DurationText:SetText(L["Time"]..": |cffffff82"..string_format("%d:%02d", duration/60, math_fmod(duration, 60))..flightEstimate.."|r"..flightHops.." - "..flightCost)
	end
end
-- ---------------------------------------------------------------------------------------------------------------------

-- ---------------------------------------------------------------------------------------------------------------------
local function SetFrameHeight(height)
	local hmf = height + height_baseheadline-1 + height_baseselection-1 + height_basestatus-1
	GVAR.MainFrame:SetHeight(hmf)
	GVAR.DataFrame:SetHeight(height)
end

local function SetFrameWidth(width)
	GVAR.MainFrame:SetWidth(width)
	GVAR.HeadlineFrame:SetWidth(width)
	GVAR.SelectionFrame:SetWidth(width)
	GVAR.SelectionFrame.ContinentButton[1]:SetPoint("TOPLEFT", GVAR.SelectionFrame, "TOPLEFT", (width - ((#continentID*32)+((#continentID-1)*width_SelectionButtonSpace))) / 2, -(5+heightFontNormal+5))

	GVAR.DataFrame:SetWidth(width)
	GVAR.DataFrame.Sort:SetWidth(width)
	for i = 1, GUI_Data_maxNum do
		GVAR.DataFrame.ScrollFrameButton[i]:SetWidth(width-offset-offset-2-16)
		GVAR.DataFrame.ScrollFrameButton[i].FlightPointButton:SetWidth(width-width_DataList_Number-width_DataList_NPC-width_DataList_Exceptions-offset-2-2-2-16-offset)
	end
	GVAR.DataFrame.Sort.FlightPointSort:SetWidth(width-width_DataList_Number-width_DataList_NPC-width_DataList_Exceptions-offset-2-2-2-16-offset)
	GVAR.StatusFrame:SetWidth(width)
end

local function CalculateFrameSize(width, height)
	local height_data = offset+(1*12)+2+(GUI_Data_curNum*12)+offset
	local height = height - height_baseheadline+1 - height_baseselection+1 - height_basestatus+1
	if height < height_data then
		local diff = height_data - height
		local freq = math_floor(diff/12)
		if freq > 0 then
			GUI_Data_curNum = GUI_Data_curNum - freq
			FlightMaster:FillList()
			height = offset+(1*12)+2+(GUI_Data_curNum*12)+offset
		end
	else
		local diff = height - height_data
		local freq = math_floor(diff/12)
		if freq > 0 then
			GUI_Data_curNum = GUI_Data_curNum + freq
			FlightMaster:FillList()
			height = offset+(1*12)+2+(GUI_Data_curNum*12)+offset
		end
	end
	SetFrameHeight(height)
	SetFrameWidth(width)
	GVAR.MainFrame:SetClampRectInsets((width-150)/2, -((width-150)/2), 0, height-1-1-1 + height_selection-1 + height_basestatus-1)
end

local function ResizeCheck(frame)
	local width = math_floor(frame:GetWidth() + 0.5)
	local height = math_floor(frame:GetHeight() + 0.5)
	CalculateFrameSize(width, height)
end

local function ResizeEnd(frame)
	frame:SetScript("OnUpdate", nil)
	frame:StopMovingOrSizing()

	local height = offset+(1*12)+2+(GUI_Data_curNum*12)+offset
	local width = math_floor(frame:GetWidth() + 0.5)

	SetFrameHeight(height)
	SetFrameWidth(width)

	GVAR.MainFrame:SetClampRectInsets((width-150)/2, -((width-150)/2), 0, height-1-1-1 + height_selection-1 + height_basestatus-1)

	FlightMaster:SaveMainFramePos()
end

local function ResizeStart(frame, point)
	frame:SetScript("OnUpdate", function() ResizeCheck(frame) end)
	frame:StartSizing(point)
end
-- ---------------------------------------------------------------------------------------------------------------------



-- ---------------------------------------------------------------------------------------------------------------------
function FlightMaster:CreateMainFrame()
	-- TaxiRouteMap Button START --------------------
	GVAR.TaxiFrameHighlightButton = CreateFrame("Button", nil, TaxiRouteMap)
	GVAR.TaxiFrameHighlightButton:SetWidth(32)
	GVAR.TaxiFrameHighlightButton:SetHeight(32)
	GVAR.TaxiFrameHighlightButton:Hide()

	GVAR.TaxiFrameHighlightButtonTexture = GVAR.TaxiFrameHighlightButton:CreateTexture(nil, "ARTWORK")
	GVAR.TaxiFrameHighlightButtonTexture:SetWidth(32)
	GVAR.TaxiFrameHighlightButtonTexture:SetHeight(32)
	GVAR.TaxiFrameHighlightButtonTexture:ClearAllPoints()
	GVAR.TaxiFrameHighlightButtonTexture:SetAllPoints(GVAR.TaxiFrameHighlightButton)
	GVAR.TaxiFrameHighlightButtonTexture:SetTexture("Interface\\PlayerFrame\\Deathknight-Energize-White")

	-- option 1 -----
	GVAR.TaxiFrameCorrectFrameLevel = CreateFrame("CheckButton", nil, TaxiRouteMap)
	TEMPLATE.CheckButton(GVAR.TaxiFrameCorrectFrameLevel, 14, 4, L["Correct FrameLevel"])
	GVAR.TaxiFrameCorrectFrameLevel:SetPoint("BOTTOMLEFT", TaxiRouteMap, "BOTTOMRIGHT", -10, 10)
	GVAR.TaxiFrameCorrectFrameLevel:SetScript("OnClick", function()
		FlightMaster_Options.CorrectFL = not FlightMaster_Options.CorrectFL
		FlightMaster:CorrectFrameLevel()
	end)
	if FlightMaster_Options.CorrectFL then
		GVAR.TaxiFrameCorrectFrameLevel:SetChecked(true)
	else
		GVAR.TaxiFrameCorrectFrameLevel:SetChecked(false)
	end
	-- option 1 -----

	-- option 2 -----
	GVAR.TaxiFrameShowOverview = CreateFrame("CheckButton", nil, TaxiRouteMap)
	TEMPLATE.CheckButton(GVAR.TaxiFrameShowOverview, 14, 4, L["Show overview"])
	GVAR.TaxiFrameShowOverview:SetPoint("BOTTOMLEFT", GVAR.TaxiFrameCorrectFrameLevel, "TOPLEFT", 0, 5)
	GVAR.TaxiFrameShowOverview:SetScript("OnClick", function()
		FlightMaster_Options.Overview = not FlightMaster_Options.Overview
		if FlightMaster_Options.Overview then
			GVAR.MainFrame:Show()
		else
			GVAR.MainFrame:Hide()
		end
	end)
	if FlightMaster_Options.Overview then
		GVAR.TaxiFrameShowOverview:SetChecked(true)
	else
		GVAR.TaxiFrameShowOverview:SetChecked(false)
	end
	-- option 2 -----

	-- option 3 -----
	GVAR.TaxiFrameShowMissing = CreateFrame("CheckButton", nil, TaxiRouteMap)
	TEMPLATE.CheckButton(GVAR.TaxiFrameShowMissing, 14, 4, L["Show missing flight points"])
	GVAR.TaxiFrameShowMissing:SetPoint("BOTTOMLEFT", GVAR.TaxiFrameShowOverview, "TOPLEFT", 0, 5)
	GVAR.TaxiFrameShowMissing:SetScript("OnClick", function()
		FlightMaster_Options.MissingFP = not FlightMaster_Options.MissingFP
		FlightMaster:MissingNode()
	end)
	if FlightMaster_Options.MissingFP then
		GVAR.TaxiFrameShowMissing:SetChecked(true)
	else
		GVAR.TaxiFrameShowMissing:SetChecked(false)
	end

	GVAR.TaxiFrameOptions = TaxiRouteMap:CreateFontString(nil, "ARTWORK", "GameFontNormal")
	GVAR.TaxiFrameOptions:SetPoint("BOTTOMLEFT", GVAR.TaxiFrameShowMissing, "TOPLEFT", 0, 5)
	GVAR.TaxiFrameOptions:SetJustifyH("LEFT")
	GVAR.TaxiFrameOptions:SetText("FlightMaster:")
	GVAR.TaxiFrameOptions:SetTextColor(0.93, 0.85, 0.22, 1)
	-- option 3 -----
	
	local minw = GVAR.TaxiFrameOptions:GetStringWidth()
	local wa   = GVAR.TaxiFrameCorrectFrameLevel.Text:GetStringWidth()
	local wb   = GVAR.TaxiFrameShowOverview.Text:GetStringWidth()
	local wc   = GVAR.TaxiFrameShowMissing.Text:GetStringWidth()

	if wa > minw then minw = wa end
	if wb > minw then minw = wb end
	if wc > minw then minw = wc end
	GVAR.TaxiFrameCorrectFrameLevel:SetPoint("BOTTOMLEFT", TaxiRouteMap, "BOTTOMRIGHT", -(minw+18+20), 10)
	-- TaxiRouteMap END --------------------



	-- Main Frame START --------------------
	GVAR.MainFrame = CreateFrame("Frame", "FlightMaster_MainFrame", UIParent)
	TEMPLATE.BorderTRBL(GVAR.MainFrame)
	GVAR.MainFrame:EnableMouse(true)
	GVAR.MainFrame:SetMovable(true)
	GVAR.MainFrame:SetResizable(true)
	GVAR.MainFrame:SetMinResize(350, height_mainframeMin) -- (minWidth, minHeight)
	GVAR.MainFrame:SetMaxResize(600, height_mainframeMax) -- (maxWidth, maxHeight)
	GVAR.MainFrame:SetToplevel(true)
	GVAR.MainFrame:SetClampedToScreen(true)
	GVAR.MainFrame:SetClampRectInsets((width_default-150)/2, -((width_default-150)/2), 0, height_data-1-1-1 + height_selection-1 + height_basestatus-1)
	GVAR.MainFrame:SetWidth(width_default)
	GVAR.MainFrame:SetHeight(height_mainframe)
	GVAR.MainFrame:SetScript("OnShow", function() FlightMaster:MainFrameShow() end)
	GVAR.MainFrame:SetScript("OnHide", function() FlightMaster:MainFrameHide() end)
	GVAR.MainFrame:Hide()



	GVAR.HeadlineFrame = CreateFrame("Frame", nil, GVAR.MainFrame)
	TEMPLATE.BorderTRBL(GVAR.HeadlineFrame)
	GVAR.HeadlineFrame:SetWidth(width_default)
	GVAR.HeadlineFrame:SetHeight(height_baseheadline)
	GVAR.HeadlineFrame:SetPoint("TOPLEFT", GVAR.MainFrame, "TOPLEFT", 0, 0)
	GVAR.HeadlineFrame:EnableMouse(true)

	GVAR.HeadlineFrame.TitleTexture = GVAR.HeadlineFrame:CreateTexture(nil, "ARTWORK")
	GVAR.HeadlineFrame.TitleTexture:SetWidth(16)
	GVAR.HeadlineFrame.TitleTexture:SetHeight(16)
	GVAR.HeadlineFrame.TitleTexture:SetPoint("LEFT", GVAR.HeadlineFrame, "LEFT", 4, 0)
	GVAR.HeadlineFrame.TitleTexture:SetTexture(AddonIcon)

	GVAR.HeadlineFrame.Title = GVAR.HeadlineFrame:CreateFontString(nil, "ARTWORK", "GameFontNormal")
	GVAR.HeadlineFrame.Title:SetPoint("LEFT", GVAR.HeadlineFrame.TitleTexture, "RIGHT", 4, 0)
	GVAR.HeadlineFrame.Title:SetJustifyH("LEFT")
	GVAR.HeadlineFrame.Title:SetText("FlightMaster")
	GVAR.HeadlineFrame.Title:SetTextColor(0.93, 0.85, 0.22, 1)

	GVAR.HeadlineFrame.Close = CreateFrame("Button", nil, GVAR.HeadlineFrame)
	TEMPLATE.IconButton(GVAR.HeadlineFrame.Close, 1)
	GVAR.HeadlineFrame.Close:SetWidth(20)
	GVAR.HeadlineFrame.Close:SetHeight(20)
	GVAR.HeadlineFrame.Close:SetPoint("RIGHT", GVAR.HeadlineFrame, "RIGHT", -4, 0)
	GVAR.HeadlineFrame.Close:SetScript("OnClick", function() FlightMaster:MainFrameToggle() end)

	GVAR.HeadlineFrame.Options = CreateFrame("Button", nil, GVAR.HeadlineFrame)
	TEMPLATE.IconButton(GVAR.HeadlineFrame.Options, 2)
	GVAR.HeadlineFrame.Options:SetWidth(20)
	GVAR.HeadlineFrame.Options:SetHeight(20)
	GVAR.HeadlineFrame.Options:SetPoint("RIGHT", GVAR.HeadlineFrame.Close, "LEFT", -4, 0)
	GVAR.HeadlineFrame.Options:SetScript("OnClick", function() FlightMaster:MainFrameToggle() InterfaceOptionsFrame_OpenToCategory("FlightMaster") end)

	GVAR.HeadlineFrame.Mover = CreateFrame("Button", nil, GVAR.HeadlineFrame)
	GVAR.HeadlineFrame.Mover:SetWidth(22)
	GVAR.HeadlineFrame.Mover:SetHeight(height_baseheadline)
	GVAR.HeadlineFrame.Mover:SetPoint("LEFT", GVAR.HeadlineFrame, "LEFT", 0, 0)
	GVAR.HeadlineFrame.Mover:SetPoint("RIGHT", GVAR.HeadlineFrame.Options, "LEFT", 0, 0)
	GVAR.HeadlineFrame.MoverTexture = GVAR.HeadlineFrame.Mover:CreateTexture(nil, "BACKGROUND")
	GVAR.HeadlineFrame.MoverTexture:SetWidth(1)
	GVAR.HeadlineFrame.MoverTexture:SetHeight(1)
	GVAR.HeadlineFrame.MoverTexture:SetPoint("LEFT", GVAR.HeadlineFrame.Title, "RIGHT", 0, 0)
	GVAR.HeadlineFrame.MoverTexture:SetPoint("TOPRIGHT", GVAR.HeadlineFrame.Mover, "TOPRIGHT", 0, 0)
	GVAR.HeadlineFrame.MoverTexture:SetPoint("BOTTOMRIGHT", GVAR.HeadlineFrame.Mover, "BOTTOMRIGHT", 0, 0)
	GVAR.HeadlineFrame.MoverTexture:SetTexture(1, 1, 1, 0)
	GVAR.HeadlineFrame.Mover:SetScript("OnEnter", function() FlightMaster:MoverInfo("MOVE", "TOP") end)
	GVAR.HeadlineFrame.Mover:SetScript("OnLeave", function() FlightMaster:MoverInfo() end)
	GVAR.HeadlineFrame.Mover:SetScript("OnMouseDown", function() GVAR.MainFrame:StartMoving() end)
	GVAR.HeadlineFrame.Mover:SetScript("OnMouseUp", function() GVAR.MainFrame:StopMovingOrSizing() FlightMaster:SaveMainFramePos() end)



	GVAR.SelectionFrame = CreateFrame("Frame", nil, GVAR.MainFrame)
	TEMPLATE.BorderTRBL(GVAR.SelectionFrame)
	GVAR.SelectionFrame:SetWidth(width_default)
	GVAR.SelectionFrame:SetHeight(height_baseselection)
	GVAR.SelectionFrame:SetPoint("TOPLEFT", GVAR.HeadlineFrame, "BOTTOMLEFT", 0, 1)
	GVAR.SelectionFrame:EnableMouse(true)



	--GVAR.SelectionFrame.MapOptions = CreateFrame("Button", nil, GVAR.SelectionFrame)
	--TEMPLATE.TextButton(GVAR.SelectionFrame.MapOptions, "Options", 4)
	--GVAR.SelectionFrame.MapOptions:SetWidth(90)
	--GVAR.SelectionFrame.MapOptions:SetHeight(20)
	--GVAR.SelectionFrame.MapOptions:SetPoint("BOTTOMLEFT", GVAR.SelectionFrame, "BOTTOMLEFT", 4, 4)
	--GVAR.SelectionFrame.MapOptions:SetScript("OnClick", function() end)



	GVAR.SelectionFrame.ContinentText = GVAR.SelectionFrame:CreateFontString(nil, "ARTWORK", "GameFontNormal")
	GVAR.SelectionFrame.ContinentText:SetPoint("TOP", GVAR.SelectionFrame, "TOP", 0, -5)
	GVAR.SelectionFrame.ContinentText:SetJustifyH("CENTER")
	GVAR.SelectionFrame.ContinentText:SetText("FlightMaster")
	GVAR.SelectionFrame.ContinentText:SetTextColor(0.93, 0.85, 0.22, 1)

	for i = 1, #continentID do
		if not GVAR.SelectionFrame.ContinentButton then GVAR.SelectionFrame.ContinentButton = {} end
		GVAR.SelectionFrame.ContinentButton[i] = CreateFrame("Button", nil, GVAR.SelectionFrame)
		GVAR.SelectionFrame.ContinentButton[i].nameMapID = continentID[i].nameID
		GVAR.SelectionFrame.ContinentButton[i]:SetWidth(32)
		GVAR.SelectionFrame.ContinentButton[i]:SetHeight(32)
		if i == 1 then
			GVAR.SelectionFrame.ContinentButton[i]:SetPoint("TOPLEFT", GVAR.SelectionFrame, "TOPLEFT", (width_default - ((#continentID*32)+((#continentID-1)*width_SelectionButtonSpace))) / 2, -(5+heightFontNormal+5))
		else
			GVAR.SelectionFrame.ContinentButton[i]:SetPoint("LEFT", GVAR.SelectionFrame.ContinentButton[i-1], "RIGHT", width_SelectionButtonSpace, 0)
		end
		GVAR.SelectionFrame.ContinentButton[i].Icon = GVAR.SelectionFrame.ContinentButton[i]:CreateTexture(nil, "BACKGROUND")
		GVAR.SelectionFrame.ContinentButton[i].Icon:SetWidth(32)
		GVAR.SelectionFrame.ContinentButton[i].Icon:SetHeight(32)
		GVAR.SelectionFrame.ContinentButton[i].Icon:SetAllPoints(GVAR.SelectionFrame.ContinentButton[i])
		GVAR.SelectionFrame.ContinentButton[i].Icon:SetTexture(continentID[i].icon)
		GVAR.SelectionFrame.ContinentButton[i].IconBorder = GVAR.SelectionFrame.ContinentButton[i]:CreateTexture(nil, "ARTWORK")
		GVAR.SelectionFrame.ContinentButton[i].IconBorder:SetWidth(64)
		GVAR.SelectionFrame.ContinentButton[i].IconBorder:SetHeight(64)
		GVAR.SelectionFrame.ContinentButton[i].IconBorder:SetPoint("TOPLEFT", -14, 14)
		GVAR.SelectionFrame.ContinentButton[i].IconBorder:SetPoint("BOTTOMRIGHT", 14, -14)
		GVAR.SelectionFrame.ContinentButton[i].IconBorder:SetTexture("Interface\\Buttons\\CheckButtonGlow")
		
		GVAR.SelectionFrame.ContinentButton[i]:SetScript("OnClick", function()
			if GUI_Selection == GVAR.SelectionFrame.ContinentButton[i].nameMapID then return end
			GUI_Selection = GVAR.SelectionFrame.ContinentButton[i].nameMapID
			FlightMaster:UpdateData()
		end)

		GVAR.SelectionFrame.ContinentButton[i]:SetScript("OnEnter", function()
			GVAR.SelectionFrame.ContinentButton[i].Icon:SetAlpha(1)
		end)

		GVAR.SelectionFrame.ContinentButton[i]:SetScript("OnLeave", function()
			if GUI_Selection == GVAR.SelectionFrame.ContinentButton[i].nameMapID then return end
			GVAR.SelectionFrame.ContinentButton[i].Icon:SetAlpha(0.7)
		end)

		GVAR.SelectionFrame.ContinentButton[i].TextTotal = GVAR.SelectionFrame.ContinentButton[i]:CreateFontString(nil, "ARTWORK", "GameFontNormalSmall")
		GVAR.SelectionFrame.ContinentButton[i].TextTotal:SetPoint("TOP", GVAR.SelectionFrame.ContinentButton[i].Icon, "BOTTOM", 0, -5)
		GVAR.SelectionFrame.ContinentButton[i].TextTotal:SetJustifyH("CENTER")
		GVAR.SelectionFrame.ContinentButton[i].TextTotal:SetTextColor(0.6, 0.6, 0.6, 1)
	end

	FlightMaster:UpdateSelectionButtons()

	GVAR.SelectionFrame.SearchBox = CreateFrame("Button", nil, GVAR.SelectionFrame)
	TEMPLATE.EditBox(GVAR.SelectionFrame.SearchBox, 150, true, FlightMaster.Search, FlightMaster.SearchClear)
	GVAR.SelectionFrame.SearchBox:SetWidth(150)
	GVAR.SelectionFrame.SearchBox:SetHeight(17)
	GVAR.SelectionFrame.SearchBox:SetPoint("TOP", GVAR.SelectionFrame, "TOP", 0, -(5+heightFontNormal+5+32+5+heightFontSmall+5))



	GVAR.DataFrame = CreateFrame("Frame", nil, GVAR.MainFrame)
	TEMPLATE.BorderTRBL(GVAR.DataFrame)
	GVAR.DataFrame:SetWidth(width_default)
	GVAR.DataFrame:SetHeight(offset+(1*12)+2+(GUI_Data_curNum*12)+offset) -- 30 = 382
	GVAR.DataFrame:SetPoint("TOPLEFT", GVAR.SelectionFrame, "BOTTOMLEFT", 0, 1)
	GVAR.DataFrame:EnableMouse(true)
	GVAR.DataFrame:EnableMouseWheel(true)
	GVAR.DataFrame:SetScript("OnMouseWheel", NOOP)

	GVAR.DataFrame.Sort = CreateFrame("Frame", nil, GVAR.DataFrame)
	TEMPLATE.DataList(GVAR.DataFrame.Sort, true)
	GVAR.DataFrame.Sort:SetWidth(width_default)
	GVAR.DataFrame.Sort:SetHeight(12)
	GVAR.DataFrame.Sort:SetPoint("TOPLEFT", GVAR.DataFrame, "TOPLEFT", offset, -offset)

	GVAR.DataFrame.ScrollFrame = CreateFrame("ScrollFrame", nil, GVAR.DataFrame)
	TEMPLATE.ScrollFrame(GVAR.DataFrame.ScrollFrame)
	GVAR.DataFrame.ScrollFrame:SetPoint("TOPLEFT", GVAR.DataFrame, "TOPLEFT", offset, -(12+2+offset))
	GVAR.DataFrame.ScrollFrame:SetPoint("BOTTOMRIGHT", GVAR.DataFrame, "BOTTOMRIGHT", -(16+offset+2), offset)
	GVAR.DataFrame.ScrollFrame:SetScript("OnVerticalScroll", function(self, value) FlightMaster:OnVerticalScroll(value) end)

	for i = 1, GUI_Data_maxNum do
		if not GVAR.DataFrame.ScrollFrameButton then GVAR.DataFrame.ScrollFrameButton = {} end
		GVAR.DataFrame.ScrollFrameButton[i] = CreateFrame("Button", nil, GVAR.DataFrame)
		TEMPLATE.DataList(GVAR.DataFrame.ScrollFrameButton[i])
		GVAR.DataFrame.ScrollFrameButton[i]:SetWidth(width_default-offset-offset-16-2)
		GVAR.DataFrame.ScrollFrameButton[i]:SetHeight(12)
		if i == 1 then
			GVAR.DataFrame.ScrollFrameButton[i]:SetPoint("TOPLEFT", GVAR.DataFrame, "TOPLEFT", offset, -(12+2+offset))
		else
			GVAR.DataFrame.ScrollFrameButton[i]:SetPoint("TOPLEFT", GVAR.DataFrame.ScrollFrameButton[i-1], "BOTTOMLEFT", 0, 0)
		end
	end



	GVAR.StatusFrame = CreateFrame("Frame", nil, GVAR.MainFrame)
	TEMPLATE.BorderTRBL(GVAR.StatusFrame)
	GVAR.StatusFrame:SetWidth(width_default)
	GVAR.StatusFrame:SetHeight(height_basestatus)
	GVAR.StatusFrame:SetPoint("TOPLEFT", GVAR.DataFrame, "BOTTOMLEFT", 0, 1)
	GVAR.StatusFrame:EnableMouse(true)
	GVAR.StatusFrame:SetScript("OnEnter", function() FlightMaster:MoverInfo("MOVE", "BOTTOM") end)
	GVAR.StatusFrame:SetScript("OnLeave", function() FlightMaster:MoverInfo() end)
	GVAR.StatusFrame:SetScript("OnMouseDown", function() GVAR.MainFrame:StartMoving() end)
	GVAR.StatusFrame:SetScript("OnMouseUp", function() GVAR.MainFrame:StopMovingOrSizing() FlightMaster:SaveMainFramePos() end)

	GVAR.StatusFrameText = GVAR.StatusFrame:CreateFontString(nil, "ARTWORK", "GameFontNormalSmall")
	GVAR.StatusFrameText:SetHeight(20)
	GVAR.StatusFrameText:SetPoint("CENTER", GVAR.StatusFrame)
	GVAR.StatusFrameText:SetJustifyH("CENTER")
	GVAR.StatusFrameText:SetTextColor(1, 1, 1, 1)

	GVAR.StatusFrame.ResizeBL = CreateFrame("Button", nil, GVAR.StatusFrame)
	GVAR.StatusFrame.ResizeBL:SetPoint("BOTTOMLEFT", GVAR.StatusFrame, "BOTTOMLEFT", 0, 0)
	GVAR.StatusFrame.ResizeBL:SetWidth(20)
	GVAR.StatusFrame.ResizeBL:SetHeight(20)

	GVAR.StatusFrame.ResizeBLtexture = GVAR.StatusFrame.ResizeBL:CreateTexture(nil, "ARTWORK")
	GVAR.StatusFrame.ResizeBLtexture:SetPoint("TOPLEFT", 2, -2)
	GVAR.StatusFrame.ResizeBLtexture:SetPoint("BOTTOMRIGHT", -2, 2)
	GVAR.StatusFrame.ResizeBLtexture:SetAlpha(0.6)
	GVAR.StatusFrame.ResizeBLtexture:SetTexture("Interface\\Minimap\\Rotating-MinimapGroupArrow")
	GVAR.StatusFrame.ResizeBLtexture:SetTexCoord(0.78125,0.484375 , 0.515625,0.21875 , 0.515625,0.75 , 0.25,0.484375) -- magic

	GVAR.StatusFrame.ResizeBL:SetScript("OnMouseDown", function()
		GVAR.StatusFrame.ResizeBLtexture:SetPoint("TOPLEFT", 3, -3)
		GVAR.StatusFrame.ResizeBLtexture:SetPoint("BOTTOMRIGHT", -3, 3)
		ResizeStart(GVAR.MainFrame, "BOTTOMLEFT")
	end)
	GVAR.StatusFrame.ResizeBL:SetScript("OnMouseUp", function()
		GVAR.StatusFrame.ResizeBLtexture:SetPoint("TOPLEFT", 2, -2)
		GVAR.StatusFrame.ResizeBLtexture:SetPoint("BOTTOMRIGHT", -2, 2)
		ResizeEnd(GVAR.MainFrame)
	end)
	GVAR.StatusFrame.ResizeBL:SetScript("OnEnter", function()
		FlightMaster:MoverInfo("RESIZE", "BOTTOM")
		GVAR.StatusFrame.ResizeBLtexture:SetAlpha(1)
	end)
	GVAR.StatusFrame.ResizeBL:SetScript("OnLeave", function()
		FlightMaster:MoverInfo()
		GVAR.StatusFrame.ResizeBLtexture:SetAlpha(0.6)
	end)

	GVAR.StatusFrame.ResizeBR = CreateFrame("Button", nil, GVAR.StatusFrame)
	GVAR.StatusFrame.ResizeBR:SetPoint("BOTTOMRIGHT", GVAR.StatusFrame, "BOTTOMRIGHT", 0, 0)
	GVAR.StatusFrame.ResizeBR:SetWidth(20)
	GVAR.StatusFrame.ResizeBR:SetHeight(20)

	GVAR.StatusFrame.ResizeBRtexture = GVAR.StatusFrame.ResizeBR:CreateTexture(nil, "ARTWORK")
	GVAR.StatusFrame.ResizeBRtexture:SetPoint("TOPLEFT", 2, -2)
	GVAR.StatusFrame.ResizeBRtexture:SetPoint("BOTTOMRIGHT", -2, 2)
	GVAR.StatusFrame.ResizeBRtexture:SetAlpha(0.6)
	GVAR.StatusFrame.ResizeBRtexture:SetTexture("Interface\\Minimap\\Rotating-MinimapGroupArrow")
	GVAR.StatusFrame.ResizeBRtexture:SetTexCoord(0.515625,0.75 , 0.78125,0.484375 , 0.25,0.484375 , 0.515625,0.21875) -- magic

	GVAR.StatusFrame.ResizeBR:SetScript("OnMouseDown", function()
		GVAR.StatusFrame.ResizeBRtexture:SetPoint("TOPLEFT", 3, -3)
		GVAR.StatusFrame.ResizeBRtexture:SetPoint("BOTTOMRIGHT", -3, 3)
		ResizeStart(GVAR.MainFrame, "BOTTOMRIGHT")
	end)
	GVAR.StatusFrame.ResizeBR:SetScript("OnMouseUp", function()
		GVAR.StatusFrame.ResizeBRtexture:SetPoint("TOPLEFT", 2, -2)
		GVAR.StatusFrame.ResizeBRtexture:SetPoint("BOTTOMRIGHT", -2, 2)
		ResizeEnd(GVAR.MainFrame)
	end)
	GVAR.StatusFrame.ResizeBR:SetScript("OnEnter", function()
		FlightMaster:MoverInfo("RESIZE", "BOTTOM")
		GVAR.StatusFrame.ResizeBRtexture:SetAlpha(1)
	end)
	GVAR.StatusFrame.ResizeBR:SetScript("OnLeave", function()
		FlightMaster:MoverInfo()
		GVAR.StatusFrame.ResizeBRtexture:SetAlpha(0.6)
	end)



	GVAR.MoverInfoTopFrame = CreateFrame("Frame", nil, GVAR.MainFrame)
	GVAR.MoverInfoTopFrame:SetWidth(150)
	GVAR.MoverInfoTopFrame:SetHeight(14)
	GVAR.MoverInfoTopFrame:SetPoint("BOTTOM", GVAR.MainFrame, "TOP", 0, 0)
	GVAR.MoverInfoTopFrame:EnableMouse(true)
	GVAR.MoverInfoTopFrame:Hide()
	GVAR.MoverInfoTopFrameTexture = GVAR.MoverInfoTopFrame:CreateTexture(nil, "BACKGROUND")
	GVAR.MoverInfoTopFrameTexture:SetAllPoints(GVAR.MoverInfoTopFrame)
	GVAR.MoverInfoTopFrameTexture:SetTexture(0, 0, 0, 0.5)
	GVAR.MoverInfoTopFrameText = GVAR.MoverInfoTopFrame:CreateFontString(nil, "ARTWORK", "GameFontNormalSmall")
	GVAR.MoverInfoTopFrameText:SetPoint("CENTER", GVAR.MoverInfoTopFrame, "CENTER", 0, 0)
	GVAR.MoverInfoTopFrameText:SetJustifyH("CENTER")
	GVAR.MoverInfoTopFrameText:SetTextColor(0.8, 0.8, 0.8, 1)

	GVAR.MoverInfoBottomFrame = CreateFrame("Frame", nil, GVAR.MainFrame)
	GVAR.MoverInfoBottomFrame:SetWidth(150)
	GVAR.MoverInfoBottomFrame:SetHeight(14)
	GVAR.MoverInfoBottomFrame:SetPoint("TOP", GVAR.StatusFrame, "BOTTOM", 0, 0)
	GVAR.MoverInfoBottomFrame:EnableMouse(true)
	GVAR.MoverInfoBottomFrame:Hide()
	GVAR.MoverInfoBottomFrameTexture = GVAR.MoverInfoBottomFrame:CreateTexture(nil, "BACKGROUND")
	GVAR.MoverInfoBottomFrameTexture:SetAllPoints(GVAR.MoverInfoBottomFrame)
	GVAR.MoverInfoBottomFrameTexture:SetTexture(0, 0, 0, 0.5)
	GVAR.MoverInfoBottomFrameText = GVAR.MoverInfoBottomFrame:CreateFontString(nil, "ARTWORK", "GameFontNormalSmall")
	GVAR.MoverInfoBottomFrameText:SetPoint("CENTER", GVAR.MoverInfoBottomFrame, "CENTER", 0, 0)
	GVAR.MoverInfoBottomFrameText:SetJustifyH("CENTER")
	GVAR.MoverInfoBottomFrameText:SetTextColor(0.8, 0.8, 0.8, 1)



	GVAR.ExceptionFrame = CreateFrame("Frame", nil, GVAR.MainFrame)
	TEMPLATE.BorderTRBL(GVAR.ExceptionFrame)
	GVAR.ExceptionFrame.FrameBackground:SetTexture(0.8, 0.2, 0.2, 1)
	GVAR.ExceptionFrame:SetToplevel(true)
	GVAR.ExceptionFrame:SetHeight(40)
	GVAR.ExceptionFrame:Hide()
	GVAR.ExceptionFrameText = GVAR.ExceptionFrame:CreateFontString(nil, "ARTWORK", "GameFontNormal")
	GVAR.ExceptionFrameText:SetPoint("CENTER")
	GVAR.ExceptionFrameText:SetTextColor(1, 1, 0.49, 1)
	-- Main Frame END --------------------



	-- Bar Frame START --------------------
	GVAR.BarFrame = CreateFrame("Frame", nil, UIParent)
	TEMPLATE.BorderTRBL(GVAR.BarFrame)
	GVAR.BarFrame:EnableMouse(true)
	GVAR.BarFrame:SetMovable(true)
	GVAR.BarFrame:SetToplevel(true)
	GVAR.BarFrame:SetClampedToScreen(true)
	GVAR.BarFrame:SetWidth(barWidth)
	GVAR.BarFrame:SetHeight(barHeight)
	GVAR.BarFrame:SetScript("OnMouseDown", function() GVAR.BarFrame:StartMoving() end)
	GVAR.BarFrame:SetScript("OnMouseUp", function() GVAR.BarFrame:StopMovingOrSizing() FlightMaster:SaveBarFramePos() end)
	GVAR.BarFrame:Hide()

	GVAR.BarFrameTexture = GVAR.BarFrame:CreateTexture(nil, "BACKGROUND")
	GVAR.BarFrameTexture:ClearAllPoints()
	GVAR.BarFrameTexture:SetAllPoints(GVAR.BarFrame)
	GVAR.BarFrameTexture:SetTexture(0, 0, 0, 1)

	GVAR.BarFrame.Close = CreateFrame("Button", nil, GVAR.BarFrame)
	TEMPLATE.IconButton(GVAR.BarFrame.Close, 1)
	GVAR.BarFrame.Close:SetWidth(16)
	GVAR.BarFrame.Close:SetHeight(16)
	GVAR.BarFrame.Close:SetPoint("TOPRIGHT", GVAR.BarFrame, "TOPRIGHT", 0, 0)
	GVAR.BarFrame.Close:SetScript("OnClick", function() FlightMaster:BarFrameHide() end)

	GVAR.BarFrame.StartText = GVAR.BarFrame:CreateFontString(nil, "ARTWORK", "GameFontNormal")
	GVAR.BarFrame.StartText:SetPoint("TOP", GVAR.BarFrame, "TOP", 0, -5)
	GVAR.BarFrame.StartText:SetJustifyH("LEFT")
	GVAR.BarFrame.StartText:SetTextColor(1, 1, 1, 1)

	GVAR.BarFrame.DestinationText = GVAR.BarFrame:CreateFontString(nil, "ARTWORK", "GameFontNormal")
	GVAR.BarFrame.DestinationText:SetPoint("TOP", GVAR.BarFrame.StartText, "BOTTOM", 0, -2)
	GVAR.BarFrame.DestinationText:SetJustifyH("LEFT")
	GVAR.BarFrame.DestinationText:SetTextColor(1, 1, 1, 1)

	GVAR.BarFrame.DurationText = GVAR.BarFrame:CreateFontString(nil, "ARTWORK", "GameFontNormal")
	GVAR.BarFrame.DurationText:SetPoint("TOP", GVAR.BarFrame.DestinationText, "BOTTOM", 0, -2)
	GVAR.BarFrame.DurationText:SetJustifyH("LEFT")
	GVAR.BarFrame.DurationText:SetTextColor(1, 1, 1, 1)



	local function setchecked()
		if FlightMaster_Options.WorldView == 1 then
			GVAR.BarFrame.Normal:SetChecked(true)    GVAR.BarFrame.Normal.Text:SetTextColor(1, 1, 0.49, 1)
			GVAR.BarFrame.Energy:SetChecked(false)   GVAR.BarFrame.Energy.Text:SetTextColor(0.6, 0.6, 0.6, 1)
			GVAR.BarFrame.Panorama:SetChecked(false) GVAR.BarFrame.Panorama.Text:SetTextColor(0.6, 0.6, 0.6, 1)
		elseif FlightMaster_Options.WorldView == 2 then
			GVAR.BarFrame.Normal:SetChecked(false)   GVAR.BarFrame.Normal.Text:SetTextColor(0.6, 0.6, 0.6, 1)
			GVAR.BarFrame.Energy:SetChecked(true)    GVAR.BarFrame.Energy.Text:SetTextColor(1, 1, 0.49, 1)
			GVAR.BarFrame.Panorama:SetChecked(false) GVAR.BarFrame.Panorama.Text:SetTextColor(0.6, 0.6, 0.6, 1)
		elseif FlightMaster_Options.WorldView == 3 then
			GVAR.BarFrame.Normal:SetChecked(false)   GVAR.BarFrame.Normal.Text:SetTextColor(0.6, 0.6, 0.6, 1)
			GVAR.BarFrame.Energy:SetChecked(false)   GVAR.BarFrame.Energy.Text:SetTextColor(0.6, 0.6, 0.6, 1)
			GVAR.BarFrame.Panorama:SetChecked(true)  GVAR.BarFrame.Panorama.Text:SetTextColor(1, 1, 0.49, 1)
		end
	end

	GVAR.BarFrame.Energy = CreateFrame("CheckButton", nil, GVAR.BarFrame)
	TEMPLATE.CheckButton(GVAR.BarFrame.Energy, 16, 4, L["Energy"], {0.6, 0.6, 0.6})
	GVAR.BarFrame.Energy:SetPoint("TOP", GVAR.BarFrame.DurationText, "BOTTOM", 0, -20)
	GVAR.BarFrame.Energy:SetScript("OnClick", function()
		if FlightMaster_Options.WorldView == 2 then
			setchecked()
			return
		end
		FlightMaster_Options.WorldView = 2
		setchecked()
		FlightMaster:SetNewWorldFramePoints()
	end)

	GVAR.BarFrame.Normal = CreateFrame("CheckButton", nil, GVAR.BarFrame)
	TEMPLATE.CheckButton(GVAR.BarFrame.Normal, 16, 4, L["Normal"], {0.6, 0.6, 0.6})
	GVAR.BarFrame.Normal:SetPoint("RIGHT", GVAR.BarFrame.Energy, "LEFT", -10, 0)
	GVAR.BarFrame.Normal:SetScript("OnClick", function()
		if FlightMaster_Options.WorldView == 1 then
			setchecked()
			return
		end
		FlightMaster_Options.WorldView = 1
		setchecked()
		FlightMaster:SetNewWorldFramePoints()
	end)

	GVAR.BarFrame.Panorama = CreateFrame("CheckButton", nil, GVAR.BarFrame)
	TEMPLATE.CheckButton(GVAR.BarFrame.Panorama, 16, 4, L["Panorama"], {0.6, 0.6, 0.6})
	GVAR.BarFrame.Panorama:SetPoint("LEFT", GVAR.BarFrame.Energy, "RIGHT", 10, 0)
	GVAR.BarFrame.Panorama:SetScript("OnClick", function()
		if FlightMaster_Options.WorldView == 3 then
			setchecked()
			return
		end
		FlightMaster_Options.WorldView = 3
		setchecked()
		FlightMaster:SetNewWorldFramePoints()
	end)

	GVAR.BarFrame.EnergyBG = GVAR.BarFrame:CreateTexture(nil, "BACKGROUND")
	GVAR.BarFrame.EnergyBG:SetPoint("CENTER")
	GVAR.BarFrame.EnergyBG:SetTexture(0, 0, 0, 1)
	GVAR.BarFrame.EnergyBG:SetWidth(100000)
	GVAR.BarFrame.EnergyBG:SetHeight(100000)
	GVAR.BarFrame.EnergyBG:Hide()

	local txt = "|cffffff82"..L["Energy Saving Mode"].."|r"
	txt = txt.. "\n\n\n"
	txt = txt.. "|TInterface\\DialogFrame\\UI-Dialog-Icon-AlertNew:24|t"
	txt = txt.. "\n"
	txt = txt.. L["3D World: OFF"]
	txt = txt.. "\n"
	txt = txt.. L["User Interface: OFF"]
	txt = txt.. "\n"
	txt = txt.. L["Sound: OFF"]
	GVAR.BarFrame.EnergyBGInfo = GVAR.BarFrame:CreateFontString(nil, "ARTWORK", "GameFontNormal")
	GVAR.BarFrame.EnergyBGInfo:SetPoint("TOP", GVAR.BarFrame, "BOTTOM", 0, -20)
	GVAR.BarFrame.EnergyBGInfo:SetJustifyH("CENTER")
	GVAR.BarFrame.EnergyBGInfo:SetTextColor(0.5, 0.5, 0.5, 1)
	GVAR.BarFrame.EnergyBGInfo:SetText(txt)
	GVAR.BarFrame.EnergyBGInfo:Hide()

	setchecked()
	--FlightMaster:GetNormalWorldFramePoints() -- WORLDTEST test=uncomment
	--GVAR.BarFrame.DurationText:SetText(L["Time"]..": |cffff0000??:?? xxx |r5 - someCost") -- WORLDTEST test=uncomment
	--FlightMaster:BarFrameShow() -- WORLDTEST test=uncomment
	-- Bar Frame END --------------------
end
-- ---------------------------------------------------------------------------------------------------------------------



-- ---------------------------------------------------------------------------------------------------------------------
function FlightMaster:GetNormalWorldFramePoints()
	-- save normal WorldFrame Points to a table
	table_wipe(worldPoints)
	worldNumPoints = WorldFrame:GetNumPoints()
	worldHeight = WorldFrame:GetHeight()
	worldWidth = WorldFrame:GetWidth()
	--print("***", worldNumPoints, worldHeight, worldWidth)

	for i = 1, worldNumPoints do
		if not worldPoints[i] then worldPoints[i] = {} end
		worldPoints[i].point, worldPoints[i].relativeTo, worldPoints[i].relativePoint, worldPoints[i].offsetX, worldPoints[i].offsetY = WorldFrame:GetPoint(i)
		--print("***", i, "#", worldPoints[i].point, worldPoints[i].relativeTo, worldPoints[i].relativePoint, worldPoints[i].offsetX, worldPoints[i].offsetY)
		if worldPoints[i].offsetX < 0 then
			worldWidth = worldWidth + -worldPoints[i].offsetX
		elseif worldPoints[i].offsetX > 0 then
			worldWidth = worldWidth + worldPoints[i].offsetX
		end
		if worldPoints[i].offsetY < 0 then
			worldHeight = worldHeight + -worldPoints[i].offsetY
		elseif worldPoints[i].offsetY > 0 then
			worldHeight = worldHeight + worldPoints[i].offsetY
		end
	end
	--print("***", worldNumPoints, worldHeight, worldWidth)
	
	barEffScale = GVAR.BarFrame:GetEffectiveScale()
	barScale = GVAR.BarFrame:GetScale()
	--print( "xWorldView=", FlightMaster_Options.WorldView, "UIParent:", "scale:", UIParent:GetScale(), "effScale:", UIParent:GetEffectiveScale() )
	--print( "xWorldView=", FlightMaster_Options.WorldView, "GVAR.BarFrame:", "scale:", GVAR.BarFrame:GetScale(), "effScale:", GVAR.BarFrame:GetEffectiveScale() )
end

function FlightMaster:SetNewWorldFramePoints(normal)
	--print("Set.NewWorld.FramePoints", normal, FlightMaster_Options.WorldView)
	if normal or FlightMaster_Options.WorldView == 1 then
		-- Normal ----------------------------------------
		WorldFrame:ClearAllPoints()
		for i = 1, worldNumPoints do
			WorldFrame:SetPoint(worldPoints[i].point, worldPoints[i].relativeTo, worldPoints[i].relativePoint, worldPoints[i].offsetX, worldPoints[i].offsetY)
		end

		GVAR.BarFrame:SetParent(UIParent)
		GVAR.BarFrame:SetScale(1)

		barEffScale = GVAR.BarFrame:GetEffectiveScale()
		barScale = GVAR.BarFrame:GetScale()

		if worldSound ~= nil then
			if worldSound == "0" then
				SetCVar("Sound_EnableAllSound", 0)
			else
				SetCVar("Sound_EnableAllSound", 1)
			end
			worldSound = nil
		end

		GVAR.BarFrame.EnergyBG:Hide()
		GVAR.BarFrame.EnergyBGInfo:Hide()
		WorldFrame:Show()
		UIParent:Show()
		SetUIVisibility(true)
	elseif FlightMaster_Options.WorldView == 2 then
		-- Energy ----------------------------------------
		if GVAR.BarFrame:GetParent() ~= nil then
			GVAR.BarFrame:SetParent(nil)
			GVAR.BarFrame:SetScale(barEffScale)
		end

		worldSound = GetCVar("Sound_EnableAllSound")
		SetCVar("Sound_EnableAllSound", 0)

		GVAR.BarFrame.EnergyBG:Show()
		GVAR.BarFrame.EnergyBGInfo:Show()
		WorldFrame:Hide()
		UIParent:Hide()
		SetUIVisibility(false)
	elseif FlightMaster_Options.WorldView == 3 then
		-- Panorama ----------------------------------------
		if GVAR.BarFrame:GetParent() ~= nil then
			GVAR.BarFrame:SetParent(nil)
			GVAR.BarFrame:SetScale(barEffScale)
		end

		local height = (worldHeight/2) - 100
		WorldFrame:ClearAllPoints()
		WorldFrame:SetPoint("TOPLEFT", 0, -height)
		WorldFrame:SetPoint("BOTTOMRIGHT", 0, height)

		if worldSound ~= nil then
			if worldSound == "0" then
				SetCVar("Sound_EnableAllSound", 0)
			else
				SetCVar("Sound_EnableAllSound", 1)
			end
			worldSound = nil
		end

		GVAR.BarFrame.EnergyBG:Hide()
		GVAR.BarFrame.EnergyBGInfo:Hide()
		WorldFrame:Show()
		UIParent:Hide()
		SetUIVisibility(false)
	end
	--print( "WorldView=", FlightMaster_Options.WorldView, "UIParent:", "scale:", UIParent:GetScale(), "effScale:", UIParent:GetEffectiveScale() )
	--print( "WorldView=", FlightMaster_Options.WorldView, "GVAR.BarFrame:", "scale:", GVAR.BarFrame:GetScale(), "effScale:", GVAR.BarFrame:GetEffectiveScale() )
end
-- ---------------------------------------------------------------------------------------------------------------------



-- ---------------------------------------------------------------------------------------------------------------------
function FlightMaster:CreateMinimapButton()
	if not FlightMaster_Options.MinimapButton then
		if FlightMaster_MinimapButton then
			FlightMaster_MinimapButton:Hide()
		end
		return
	else
		if FlightMaster_MinimapButton then
			FlightMaster_MinimapButton:Show()
			return
		end
	end

	local function MoveMinimapButton()
		local xpos
		local ypos
		local minimapShape = GetMinimapShape and GetMinimapShape() or "ROUND"
		if minimapShape == "SQUARE" then
			xpos = 110 * cos(FlightMaster_Options.MinimapButtonPos or 0)
			ypos = 110 * sin(FlightMaster_Options.MinimapButtonPos or 0)
			xpos = math.max(-82, math.min(xpos, 84))
			ypos = math.max(-86, math.min(ypos, 82))
		else
			xpos = 80 * cos(FlightMaster_Options.MinimapButtonPos or 0)
			ypos = 80 * sin(FlightMaster_Options.MinimapButtonPos or 0)
		end
		FlightMaster_MinimapButton:SetPoint("TOPLEFT", Minimap, "TOPLEFT", 54-xpos, ypos-54)
	end

	local function DragMinimapButton()
		local xpos, ypos = GetCursorPosition()
		local xmin, ymin = Minimap:GetLeft() or 400, Minimap:GetBottom() or 400
		local scale = Minimap:GetEffectiveScale()
		xpos = xmin-xpos/scale+70
		ypos = ypos/scale-ymin-70
		FlightMaster_Options.MinimapButtonPos = math.deg(math.atan2(ypos, xpos))
		MoveMinimapButton()
	end

	local MinimapButton = CreateFrame("Button", "FlightMaster_MinimapButton", Minimap)
	MinimapButton:EnableMouse(true)
	MinimapButton:SetMovable(true)
	MinimapButton:SetToplevel(true)
	MinimapButton:SetWidth(32)
	MinimapButton:SetHeight(32)
	MinimapButton:SetPoint("TOPLEFT")
	MinimapButton:SetFrameStrata("MEDIUM")
	MinimapButton:RegisterForClicks("AnyUp")
	MinimapButton:RegisterForDrag("LeftButton")

	local texture = MinimapButton:CreateTexture(nil, "ARTWORK")
	texture:SetWidth(54)
	texture:SetHeight(54)
	texture:SetPoint("TOPLEFT")
	texture:SetTexture("Interface\\Minimap\\MiniMap-TrackingBorder")

	local texture = MinimapButton:CreateTexture(nil, "BACKGROUND")
	texture:SetWidth(24)
	texture:SetHeight(24)
	texture:SetPoint("TOPLEFT", 2, -4)
	texture:SetTexture("Interface\\Minimap\\UI-Minimap-Background")

	local NormalTexture = MinimapButton:CreateTexture(nil, "ARTWORK")
	NormalTexture:SetWidth(12)
	NormalTexture:SetHeight(14)
	NormalTexture:SetPoint("TOPLEFT", 10.5, -8.5)
	NormalTexture:SetTexture(AddonIcon)
	NormalTexture:SetTexCoord(2/16, 14/16, 1/16, 15/16)
	MinimapButton:SetNormalTexture(NormalTexture)

	local PushedTexture = MinimapButton:CreateTexture(nil, "ARTWORK")
	PushedTexture:SetWidth(10)
	PushedTexture:SetHeight(12)
	PushedTexture:SetPoint("TOPLEFT", 11.5, -9.5)
	PushedTexture:SetTexture(AddonIcon)
	PushedTexture:SetTexCoord(2/16, 14/16, 1/16, 15/16)
	MinimapButton:SetPushedTexture(PushedTexture)

	local HighlightTexture = MinimapButton:CreateTexture(nil, "ARTWORK")
	HighlightTexture:SetPoint("TOPLEFT", 0, 0)
	HighlightTexture:SetPoint("BOTTOMRIGHT", 0, 0)
	HighlightTexture:SetTexture("Interface\\Minimap\\UI-Minimap-ZoomButton-Highlight")
	MinimapButton:SetHighlightTexture(HighlightTexture)

	MinimapButton:SetScript("OnEnter", function(self)
		GameTooltip:SetOwner(self, "ANCHOR_RIGHT")
		GameTooltip:AddLine("FlightMaster", 1, 0.82, 0, 1)
		GameTooltip:Show()
	end)
	MinimapButton:SetScript("OnLeave", function(self) GameTooltip:Hide() end)
	MinimapButton:SetScript("OnClick", function(self, button) FlightMaster:MainFrameToggle() end)
	MinimapButton:SetScript("OnDragStart", function(self) self:LockHighlight() self:SetScript("OnUpdate", DragMinimapButton) end)
	MinimapButton:SetScript("OnDragStop", function(self) self:SetScript("OnUpdate", nil) self:UnlockHighlight() end)

	MoveMinimapButton()
end
-- ---------------------------------------------------------------------------------------------------------------------

-- ---------------------------------------------------------------------------------------------------------------------
function FlightMaster:BarFrameShow()
	if FlightMaster_Options.BarFrame_relP then
		GVAR.BarFrame:ClearAllPoints()
		GVAR.BarFrame:SetPoint(FlightMaster_Options.BarFrame_relP, FlightMaster_Options.BarFrame_posX, FlightMaster_Options.BarFrame_posY)
	else
		GVAR.BarFrame:ClearAllPoints()
		GVAR.BarFrame:SetPoint("CENTER", 0, 0)
	end
	GVAR.BarFrame:StartMoving()
	GVAR.BarFrame:StopMovingOrSizing()

	BarUpdate_TimerStart = GetTime()
	GVAR.BarFrame:SetScript("OnUpdate", BarUpdate) -- WORLDTEST test=comment

	local start       = FlightMaster_local_Nodes[locale][playerFaction][flightMap][flightStart] -- WORLDTEST test=comment
	local destination = FlightMaster_local_Nodes[locale][playerFaction][flightMap][flightDestination] -- WORLDTEST test=comment
	--local start = "start" -- WORLDTEST test=uncomment
	--local destination = "destination" -- WORLDTEST test=uncomment

	GVAR.BarFrame.StartText:SetText(L["Start"]..": |cffffff82"..start.."|r")
	GVAR.BarFrame.DestinationText:SetText(L["Destination"]..": |cffffff82"..destination.."|r")

	local textwidth1 = GVAR.BarFrame.StartText:GetStringWidth()+50
	local textwidth2 = GVAR.BarFrame.DestinationText:GetStringWidth()+50
	if textwidth1 > barWidth or textwidth2 > barWidth then
		if textwidth1 > textwidth2 then
			GVAR.BarFrame:SetWidth(textwidth1)
		else
			GVAR.BarFrame:SetWidth(textwidth2)
		end
	else
		GVAR.BarFrame:SetWidth(barWidth)
	end
	GVAR.BarFrame:Show()
end

function FlightMaster:BarFrameHide()
	BarUpdate_TimerDuration = nil
	GVAR.BarFrame:SetScript("OnUpdate", nil)
	GVAR.BarFrame:Hide()
	if FlightMaster_Options.WorldView > 1 then
		FlightMaster:SetNewWorldFramePoints(true)
	end
end

function FlightMaster:SaveBarFramePos()
	local point, relativeTo, relativePoint, x, y = GVAR.BarFrame:GetPoint()
	FlightMaster_Options.BarFrame_posX = x
	FlightMaster_Options.BarFrame_posY = y
	FlightMaster_Options.BarFrame_relP = relativePoint
end
-- ---------------------------------------------------------------------------------------------------------------------

-- ---------------------------------------------------------------------------------------------------------------------
function FlightMaster:InitializeFrame()
	if FlightMaster_Options.Data_SortBy then
		GUI_SortBy = FlightMaster_Options.Data_SortBy
	end
	if FlightMaster_Options.Data_SortOrder then
		GUI_SortOrder = FlightMaster_Options.Data_SortOrder
	end
	FlightMaster:UpdateSortButtons()
	if FlightMaster_Options.MainFrame_width then
		--FlightMaster:UpdateData()
		CalculateFrameSize(FlightMaster_Options.MainFrame_width, FlightMaster_Options.MainFrame_height)
	end
end

function FlightMaster:MainFrameToggle()
	if GVAR.MainFrame:IsShown() then
		GVAR.MainFrame:Hide()
	else
		GVAR.MainFrame:Show()
	end
end

function FlightMaster:MainFramePos()
	if FlightMaster_Options.MainFrame_relP then
		GVAR.MainFrame:ClearAllPoints()
		GVAR.MainFrame:SetPoint(FlightMaster_Options.MainFrame_relP, FlightMaster_Options.MainFrame_posX, FlightMaster_Options.MainFrame_posY)
	else
		GVAR.MainFrame:ClearAllPoints()
		GVAR.MainFrame:SetPoint("CENTER", 0, 0)
	end
end

function FlightMaster:MainFrameHide()
	PlaySound("igQuestListClose")
	GVAR.InterfaceOptions.GUI:Enable()
	GVAR.TaxiFrameHighlightButton:Hide()
end

function FlightMaster:MainFrameShow()
	PlaySound("igQuestListOpen")
	GVAR.InterfaceOptions.GUI:Disable()
	FlightMaster:MainFramePos()
	GVAR.MainFrame:StartMoving()
	GVAR.MainFrame:StopMovingOrSizing()
	FlightMaster:UpdateData()
	--print("MainFrameShow UpdateData")
end

function FlightMaster:SaveMainFramePos()
	local point, relativeTo, relativePoint, x, y = GVAR.MainFrame:GetPoint()
	FlightMaster_Options.MainFrame_posX = x
	FlightMaster_Options.MainFrame_posY = y
	FlightMaster_Options.MainFrame_relP = relativePoint
	FlightMaster_Options.MainFrame_width = math_floor(GVAR.MainFrame:GetWidth() + 0.5)
	FlightMaster_Options.MainFrame_height = math_floor(GVAR.MainFrame:GetHeight() + 0.5)
end

function FlightMaster:Search()
	SearchQuery = GVAR.SelectionFrame.SearchBox.EditBox:GetText()
	if SearchQuery == "" then
		SearchQuery = nil
	end
	if SearchQuery then
		GVAR.SelectionFrame.SearchBox.EditBoxClear:Enable()
	else
		GVAR.SelectionFrame.SearchBox.EditBoxClear:Disable()
	end
	FlightMaster:UpdateData()
end

function FlightMaster:SearchClear()
	SearchQuery = nil
	GVAR.SelectionFrame.SearchBox.EditBox:SetText("")
	GVAR.SelectionFrame.SearchBox.EditBox:ClearFocus()
end

function FlightMaster:MoverInfo(action, where)
	if not action then
		GVAR.MoverInfoTopFrame:Hide()
		GVAR.MoverInfoBottomFrame:Hide()
		return
	end

	if action == "MOVE" then
		GVAR.MoverInfoTopFrameText:SetText(L["click & move"])
		GVAR.MoverInfoBottomFrameText:SetText(L["click & move"])
	else
		GVAR.MoverInfoTopFrameText:SetText(L["click & resize"])
		GVAR.MoverInfoBottomFrameText:SetText(L["click & resize"])
	end

	if where == "TOP" then
		GVAR.MoverInfoTopFrame:Show()
	else
		GVAR.MoverInfoBottomFrame:Show()
	end
end
-- ---------------------------------------------------------------------------------------------------------------------



-- ---------------------------------------------------------------------------------------------------------------------
function FlightMaster:NodeHighlight(node, state)
	--print("Node Highlight", node, state)
	if TaxiFrame:IsShown() then
		if GUI_Selection == flightMap then
			if state then
				if reposNode[GUI_Selection] and reposNode[GUI_Selection][node] then
					node = reposNode[GUI_Selection][node]
				end
				local kx = math_floor((node - 1000000) / 1000) -- key 0.606
				local ky = (node - 1000000) - (kx * 1000)      -- key 0.192
				kx = kx/1000 * myTAXI_MAP_WIDTH
				ky = ky/1000 * myTAXI_MAP_HEIGHT
				GVAR.TaxiFrameHighlightButton:Show()
				GVAR.TaxiFrameHighlightButton:SetPoint("CENTER", TaxiFrame.InsetBg, "BOTTOMLEFT", math_floor(kx+0.5), math_floor(ky+0.5))
			else
				GVAR.TaxiFrameHighlightButton:Hide()
			end
		end
	end
end

function FlightMaster:ShowException(button, state)
	--print(button.buttonNum, GUI_Data[button.buttonNum].exceptionL, state) -- TEST
	if state and GUI_Data[button.buttonNum].exceptionL then
		GVAR.ExceptionFrameText:SetText( GUI_Data[button.buttonNum].exceptionL )
		GVAR.ExceptionFrame:SetWidth( GVAR.ExceptionFrameText:GetStringWidth()+50 )
		GVAR.ExceptionFrame:ClearAllPoints()
		GVAR.ExceptionFrame:SetPoint("BOTTOM", button.ExceptionsButton, "TOP", 0, 0)
		GVAR.ExceptionFrame:Show()
	else
		GVAR.ExceptionFrame:Hide()
	end
end

function FlightMaster:CorrectFrameLevel()
	if TaxiFrame:IsShown() then
		if FlightMaster_Options.CorrectFL then
			local maxFrameLevel = 1
			local numNodes = NumTaxiNodes()
			for index = 1, numNodes do
				local button = _G["TaxiButton"..index]
				if button then
					local fl = button:GetFrameLevel()
					if fl > maxFrameLevel then
						--print("maxFrameLevel: new =", fl, "old =", maxFrameLevel)
						maxFrameLevel = fl
					end
				end
			end
			--print("maxFrameLevel:", maxFrameLevel)
			for index = 1, numNodes do
				local button = _G["TaxiButton"..index]
				if button then
					button:SetFrameLevel(maxFrameLevel)
				end
			end
		end
	end
end

function FlightMaster:MissingNode()
	if not TaxiFrame:IsShown() then return end

	-- 1) hide all
	for flightMap in pairs(FlightMaster_NPC[playerFaction]) do
		for npc in pairs(FlightMaster_NPC[playerFaction][flightMap]) do
			local button = "TaxiFrameButton"..flightMap..npc
			if GVAR[button] then
				GVAR[button]:Hide()
				--print("npc:", npc, button)
			end
		end
	end

	-- 2) show all unreachable flight points ("NONE")
	-- 3) show missing flight points
	if FlightMaster_Options.MissingFP then
		if FlightMaster_NPC[playerFaction] and FlightMaster_NPC[playerFaction][flightMap] then
			for npc, node in pairs(FlightMaster_NPC[playerFaction][flightMap]) do

				if (flightNone[node] and FlightMaster_PlayerNodes[flightMap] and FlightMaster_PlayerNodes[flightMap][node]) or
				   (not FlightMaster_PlayerNodes[flightMap] or not FlightMaster_PlayerNodes[flightMap][node])
				then
					local button = "TaxiFrameButton"..flightMap..npc
					if not GVAR[button] then
						GVAR[button] = CreateFrame("Button", nil, TaxiRouteMap)
						GVAR[button]:SetWidth(16)
						GVAR[button]:SetHeight(16)

						local nodename
						if FlightMaster_local_Nodes[locale] and
						   FlightMaster_local_Nodes[locale][playerFaction] and
						   FlightMaster_local_Nodes[locale][playerFaction][flightMap] and
						   FlightMaster_local_Nodes[locale][playerFaction][flightMap][node]
						then
							nodename = FlightMaster_local_Nodes[locale][playerFaction][flightMap][node]
						end
						if not nodename then
							if FlightMaster_local_Nodes[locale] and
							   FlightMaster_local_Nodes[locale][oppositeFaction] and
							   FlightMaster_local_Nodes[locale][oppositeFaction][flightMap] and
							   FlightMaster_local_Nodes[locale][oppositeFaction][flightMap][node] then
								nodename = FlightMaster_local_Nodes[locale][oppositeFaction][flightMap][node]
							end
						end
						if not nodename then
							nodename = "* "..L["Unknown!"].." "..string_format(L["Never seen with %s."], locale).." *"
						end
					
						local _, exsum = FlightMaster:GetException(playerFaction, flightMap, npc)

						-- ----------
						GVAR[button]:SetScript("OnEnter", function()
							GameTooltip:SetOwner(GVAR[button], "ANCHOR_RIGHT")
							GameTooltip:AddLine(nodename, 1, 0.82, 0)
							GameTooltip:AddLine(L["NPC-ID"]..": "..npc, 1, 1, 1)
							if flightNone[node] then
								GameTooltip:AddLine(L["Currently unreachable!"], 0.8, 0.8, 0.8)
							else
								GameTooltip:AddLine(L["Missing flight point!"], 1, 0.2, 0.2)
							end
							if exsum then
								GameTooltip:AddLine(exsum, 1, 1, 0.49, 1)
							end
							GameTooltip:Show()
							--print("OnEnter", GVAR.MainFrame:IsShown(), GUI_Selection, flightMap, node, npc)
							GVAR.DataFrame.ScrollFrameButton.lockedHighlightButtonNum = nil
							if GVAR.MainFrame:IsShown() and GUI_Selection == flightMap then -- flightMap = GUI_Selection = 
								for j = 1, GUI_Data_maxNum do
									--print("j=", j, "max=", GUI_Data_maxNum)
									if GVAR.DataFrame.ScrollFrameButton[j].node == node then
										GVAR.DataFrame.ScrollFrameButton.lockedHighlightButtonNum = j
										GVAR.DataFrame.ScrollFrameButton[j]:LockHighlight()
										break
									end
								end
								FlightMaster:NodeHighlight(node, 1)
							end
						end)
						-- ----------
						-- ----------
						GVAR[button]:SetScript("OnLeave", function()
							GameTooltip:Hide()
							--print("OnLeave", GVAR.MainFrame:IsShown(), GUI_Selection, flightMap)
							if GVAR.MainFrame:IsShown() and GUI_Selection == flightMap then
								if GVAR.DataFrame.ScrollFrameButton.lockedHighlightButtonNum then
									GVAR.DataFrame.ScrollFrameButton[GVAR.DataFrame.ScrollFrameButton.lockedHighlightButtonNum]:UnlockHighlight()
								end
								FlightMaster:NodeHighlight()
							end
							GVAR.DataFrame.ScrollFrameButton.lockedHighlightButtonNum = nil
						end)
						-- ----------
						local buttontexture = "TaxiFrameButtonTexture"..flightMap..npc
						GVAR[buttontexture] = GVAR[button]:CreateTexture(nil, "ARTWORK")
						GVAR[buttontexture]:SetWidth(16*0.8)
						GVAR[buttontexture]:SetHeight(16*0.8)
						GVAR[buttontexture]:SetPoint("CENTER", 0, 0)
					end

					if flightNone[node] then
						GVAR["TaxiFrameButtonTexture"..flightMap..npc]:SetTexture("Interface\\TaxiFrame\\UI-Taxi-Icon-Gray")
					else
						GVAR["TaxiFrameButtonTexture"..flightMap..npc]:SetTexture("Interface\\TaxiFrame\\UI-Taxi-Icon-Red")
					end

					local posnode = node
					if reposNode[flightMap] and reposNode[flightMap][node] then
						posnode = reposNode[flightMap][node]
					end
					local kx = math_floor((posnode - 1000000) / 1000) -- key 0.606
					local ky = (posnode - 1000000) - (kx * 1000)      -- key 0.192
					kx = kx/1000 * myTAXI_MAP_WIDTH
					ky = ky/1000 * myTAXI_MAP_HEIGHT
					GVAR[button]:SetPoint("CENTER", TaxiFrame.InsetBg, "BOTTOMLEFT", math_floor(kx+0.5), math_floor(ky+0.5))
					GVAR[button]:Show()
					--print( flightMap, npc, button, GVAR[button]:GetFrameLevel() )
				end
			end
		end
	end

end

function FlightMaster:GetException(playerFaction, flightMap, npc)
	if FlightMaster_exNPC[playerFaction] and
	   FlightMaster_exNPC[playerFaction][flightMap] and
	   FlightMaster_exNPC[playerFaction][flightMap][npc]
	then
		local excp = FlightMaster_exNPC[playerFaction][flightMap][npc]
		local exall = string.sub(excp, string.find(excp,"#")+1 )
		local exsum = " "
		local exshort = ""
		for k, v in pairs({strsplit(":", exall)}) do
			local xx = string.sub(v, 0, 1)
			if xx == "w" then
				exsum = exsum.."("..L["Underwater"]..") "
				exshort = exshort..L["UW"].." "
			elseif xx == "c" then
				local class = string.sub(v, 3)
				exsum = exsum.."("..CLASS..": "..string_format(L["%s only"], LOCALIZED_CLASS_NAMES_MALE[class])..") "
				exshort = exshort..L["C"].." "
			elseif xx == "l" then
				local level = string.sub(v, 3)
				exsum = exsum.."("..string_format(L["Available at level %s"], level)..") "
				exshort = exshort..L["LVL"].." "
			elseif xx == "f" then
				local faction = string.sub(v, 3)
				if faction == "aldor" then
					exsum = exsum.."("..L["Aldor only"]..") "
				elseif faction == "scryers" then
					exsum = exsum.."("..L["Scryers only"]..") "
				end
				exshort = exshort..L["F"].." "
			elseif xx == "a" then
				local quest = string.sub(v, 3)
				exsum = exsum.."("..string_format(L["Available after quest id=%s"], quest)..") "
				exshort = exshort..L["+Q"].." "
			elseif xx == "u" then
				local quest = string.sub(v, 3)
				exsum = exsum.."("..string_format(L["Unavailable after quest id=%s"], quest)..") "
				exshort = exshort..L["-Q"].." "
			elseif xx == "s" then
				local npcid = string.sub(v, 3)
				exsum = exsum.."("..string_format(L["Same flight point as %s, but in a different phase."], npcid)..") "
				exshort = exshort.."== "
			end
		end
		return strtrim(exshort), strtrim(exsum)
	end
	return nil, nil
end


-- ---------------------------------------------------------------------------------------------------------------------



-- ---------------------------------------------------------------------------------------------------------------------
function FlightMaster:UpdateData()
	if not GUI_Selection then
		SetMapToCurrentZone()
		local map = GetCurrentMapContinent()
		if map > 0 and map < 5 then
			GUI_Selection = continentID[ map ].nameID
		elseif map == 6 then
			GUI_Selection = continentID[ 5 ].nameID
		else
			GUI_Selection = continentID[ 1 ].nameID -- fallback
		end
	end

	FlightMaster:UpdateSelectionButtons()

	FlightMaster:CreateGUIdata()
	FlightMaster:FillList()
	FlightMaster:UpdateStatus()
end

function FlightMaster:UpdateSelectionButtons()
	for i = 1, #continentID do
		if GUI_Selection == GVAR.SelectionFrame.ContinentButton[i].nameMapID then
			GVAR.SelectionFrame.ContinentButton[i].IconBorder:Show()
			GVAR.SelectionFrame.ContinentButton[i].Icon:SetAlpha(1)
			GVAR.SelectionFrame.ContinentText:SetText(continentID[i].name)
		else
			GVAR.SelectionFrame.ContinentButton[i].IconBorder:Hide()
			GVAR.SelectionFrame.ContinentButton[i].Icon:SetAlpha(0.7)
		end
	end
end

function FlightMaster:UpdateStatus()
	if FlightMaster_PlayerNodes[GUI_Selection] and FlightMaster_PlayerNodes[GUI_Selection]["UPDATE"] then
		GVAR.StatusFrameText:SetText("|cffffff82"..L["Latest Update"].."|r: "..localizedTimeDate(FlightMaster_PlayerNodes[GUI_Selection]["UPDATE"], 2))
		GVAR.MainFrame.FrameBackground:SetTexture(0.4, 0.4, 0.4, 1)
		GVAR.HeadlineFrame.FrameBackground:SetTexture(0.4, 0.4, 0.4, 1)
		GVAR.SelectionFrame.FrameBackground:SetTexture(0.4, 0.4, 0.4, 1)
		GVAR.DataFrame.FrameBackground:SetTexture(0.4, 0.4, 0.4, 1)
		GVAR.StatusFrame.FrameBackground:SetTexture(0.4, 0.4, 0.4, 1)
	else
		local name = GUI_Selection
		for i = 1, #continentID do
			if GUI_Selection == continentID[i].nameID then
				name = continentID[i].name
				break
			end
		end

		GVAR.StatusFrameText:SetText("|cffffff82"..name.."|r: |cffff0000"..L["No data. Visit a flight master!"].."|r")
		GVAR.MainFrame.FrameBackground:SetTexture(0.8, 0.2, 0.2, 1)
		GVAR.HeadlineFrame.FrameBackground:SetTexture(0.8, 0.2, 0.2, 1)
		GVAR.SelectionFrame.FrameBackground:SetTexture(0.8, 0.2, 0.2, 1)
		GVAR.DataFrame.FrameBackground:SetTexture(0.8, 0.2, 0.2, 1)
		GVAR.StatusFrame.FrameBackground:SetTexture(0.8, 0.2, 0.2, 1)
	end

	-- total
	if not continentID[1].FPtotal then
		for i = 1, #continentID do
			local map = continentID[i].nameID
			continentID[i].FPtotal = 0
			for npc in pairs(FlightMaster_NPC[playerFaction][map]) do
				continentID[i].FPtotal = continentID[i].FPtotal + 1
				--print("FPtotal", i, "#", npc, continentID[i].FPtotal)
			end
		end
	end

	-- known
	for i = 1, #continentID do
		local map = continentID[i].nameID
		continentID[i].FPknown = 0
		if FlightMaster_PlayerNodes[map] then
			for node, known in pairs(FlightMaster_PlayerNodes[map]) do

				if known == 1 and FlightMaster_NPC[playerFaction] and FlightMaster_NPC[playerFaction][map] then
					for npc, NPCnode in pairs(FlightMaster_NPC[playerFaction][map]) do
						--print(i, continentID[i].FPknown, node, NPCnode)
						if node == NPCnode then
							continentID[i].FPknown = continentID[i].FPknown + 1
							--print(i, continentID[i].FPknown, node, NPCnode)
						end
					end
				end

			end
		end
	end

	-- text output
	for i = 1, #continentID do
		local map = continentID[i].nameID
		--print(i, continentID[i].FPknown, continentID[i].FPtotal)
		local color = "|cffff0000" -- red (1, 0, 0)
		if continentID[i].FPknown >= continentID[i].FPtotal then
			continentID[i].FPknown = continentID[i].FPtotal
			color = "|cff00bf00" -- green (0, 0.75, 0)
		end
		if not FlightMaster_PlayerNodes[map] then
			color = "|cff4c4c4c" -- grey (0.3, 0.3, 0.3)
			GVAR.SelectionFrame.ContinentButton[i].TextTotal:SetText(color.."??/"..continentID[i].FPtotal.."|r")
		else
			GVAR.SelectionFrame.ContinentButton[i].TextTotal:SetText(color..continentID[i].FPknown.."/"..continentID[i].FPtotal.."|r")
		end
	end
end

function FlightMaster:CreateGUIdata()
	table_wipe(GUI_Data)

	local x = 1
	for npc, node in pairs(FlightMaster_NPC[playerFaction][GUI_Selection]) do
		GUI_Data[x] = {}
		GUI_Data[x].npcid = npc
		GUI_Data[x].node = node
		if FlightMaster_local_Nodes[locale] and
		   FlightMaster_local_Nodes[locale][playerFaction] and
		   FlightMaster_local_Nodes[locale][playerFaction][GUI_Selection] and
		   FlightMaster_local_Nodes[locale][playerFaction][GUI_Selection][node] then
			GUI_Data[x].name = FlightMaster_local_Nodes[locale][playerFaction][GUI_Selection][node]
		end
		if not GUI_Data[x].name then
			if FlightMaster_local_Nodes[locale] and
			   FlightMaster_local_Nodes[locale][oppositeFaction] and
			   FlightMaster_local_Nodes[locale][oppositeFaction][GUI_Selection] and
			   FlightMaster_local_Nodes[locale][oppositeFaction][GUI_Selection][node] then
				GUI_Data[x].name = FlightMaster_local_Nodes[locale][oppositeFaction][GUI_Selection][node]
			end
		end
		if not GUI_Data[x].name then
			GUI_Data[x].name = "* "..L["Unknown!"].." "..string_format(L["Never seen with %s."], locale).." *"
		end

		local exshort, exsum = FlightMaster:GetException(playerFaction, GUI_Selection, npc)
		if exshort and exsum then
			GUI_Data[x].exception = exshort
			GUI_Data[x].exceptionL = exsum
		else
			GUI_Data[x].exception = "zzzzz"
			--GUI_Data[x].exceptionL = ""
		end

		if FlightMaster_PlayerNodes[GUI_Selection] and FlightMaster_PlayerNodes[GUI_Selection][node] then
			GUI_Data[x].known = 1
		end
		x = x + 1
	end

	GUI_Data_Count = #GUI_Data

	local sortfunc
	if GUI_SortBy == 1 then -- npcid
		if GUI_SortOrder == 1 then
			sortfunc = function(a, b)
				if a.npcid < b.npcid then return true end
			end
		else
			sortfunc = function(a, b)
				if a.npcid > b.npcid then return true end
			end
		end
	elseif GUI_SortBy == 2 then -- exception
		if GUI_SortOrder == 1 then
			sortfunc = function(a, b)
				if a.exception < b.exception then return true end
			end
		else
			sortfunc = function(a, b)
				if a.exception > b.exception then return true end
			end
		end
	elseif GUI_SortBy == 3 then -- name
		if GUI_SortOrder == 1 then
			sortfunc = function(a, b)
				if a.name < b.name then return true end
			end
		else
			sortfunc = function(a, b)
				if a.name > b.name then return true end
			end
		end
	end

	table_sort(GUI_Data, sortfunc)

	if SearchQuery then
		local newtemp = {}
		local x = 1
		local lowerSearchQuery = string_lower(SearchQuery)
		for i = 1, GUI_Data_Count do
			if string_find(string_lower(GUI_Data[i].name), lowerSearchQuery, 1, true) then
				newtemp[x] = {}
				newtemp[x].npcid      = GUI_Data[i].npcid
				newtemp[x].node       = GUI_Data[i].node
				newtemp[x].exception  = GUI_Data[i].exception
				newtemp[x].exceptionL = GUI_Data[i].exceptionL
				newtemp[x].name       = GUI_Data[i].name
				newtemp[x].known      = GUI_Data[i].known
				x = x + 1
			end
		end
		table_wipe(GUI_Data)
		for i = 1, #newtemp do
			GUI_Data[i] = newtemp[i]
		end
		table_wipe(newtemp)
		GUI_Data_Count = #GUI_Data
	end
end

function FlightMaster:GUIsort(sortBy)
	if GUI_SortBy == sortBy then
		if GUI_SortOrder == 1 then
			GUI_SortOrder = 0
		else
			GUI_SortOrder = 1
		end
	else
		GUI_SortOrder = 1
	end
	GUI_SortBy = sortBy

	FlightMaster_Options.Data_SortBy = GUI_SortBy
	FlightMaster_Options.Data_SortOrder = GUI_SortOrder

	FlightMaster:UpdateSortButtons()
	FlightMaster:CreateGUIdata()
	FlightMaster:FillList()
end

function FlightMaster:UpdateSortButtons()
	local arrow
	for i = 1, 3 do
		if i == 1 then -- NPC
			arrow = GVAR.DataFrame.Sort.NPCSort.arrow
		elseif i == 2 then -- Exceptions
			arrow = GVAR.DataFrame.Sort.ExceptionsSort.arrow
		elseif i == 3 then -- FlightPoint
			arrow = GVAR.DataFrame.Sort.FlightPointSort.arrow
		end
		if GUI_SortBy == i then
			if GUI_SortOrder == 1 then arrow:SetTexCoord(0, 0.5625, 0, 1) end
			if GUI_SortOrder == 0 then arrow:SetTexCoord(0.5625, 0, 1, 0) end
		else
			arrow:SetTexCoord(0, 0, 0, 0)
		end
	end
end

function FlightMaster:OnVerticalScroll(value)
	GVAR.DataFrame.ScrollFrame.ScrollBar:SetValue(value)
	GVAR.DataFrame.ScrollFrame.offset = math_floor((value / 12) + 0.5)
	FlightMaster:FillList()
end

function FlightMaster:FillList()
	local scrollFrame = GVAR.DataFrame.ScrollFrame
	local scrollFrameHeight = 0
	local numToDisplay = GUI_Data_curNum
	local valueStep = 12

	if GUI_Data_Count < numToDisplay then
		numToDisplay = GUI_Data_Count
	end

	if GUI_Data_Count > 0 then
		scrollFrameHeight = (GUI_Data_Count - numToDisplay) * valueStep
		if scrollFrameHeight < 0 then
			scrollFrameHeight = 0
		end
	end
	scrollFrame.ScrollBar:SetMinMaxValues(0, scrollFrameHeight)
	scrollFrame.ScrollBar:SetValueStep(valueStep)

	if GUI_Data_Count <= GUI_Data_curNum then
		scrollFrame.ScrollBarScrollUpButton:Disable()
		scrollFrame.ScrollBarScrollDownButton:Disable()
		scrollFrame.ScrollBarScrollUpButton:SetAlpha(0.2)
		scrollFrame.ScrollBarScrollDownButton:SetAlpha(0.2)
		scrollFrame.ScrollBarThumbTexture:SetAlpha(0.2)
	else
		scrollFrame.ScrollBarScrollUpButton:Enable()
		scrollFrame.ScrollBarScrollDownButton:Enable()
		scrollFrame.ScrollBarScrollUpButton:SetAlpha(1)
		scrollFrame.ScrollBarScrollDownButton:SetAlpha(1)
		scrollFrame.ScrollBarThumbTexture:SetAlpha(1)
	end

	if scrollFrame.ScrollBar:GetValue() == 0 then
		scrollFrame.ScrollBarScrollUpButton:Disable()
	else
		scrollFrame.ScrollBarScrollUpButton:Enable()
	end
	if ((scrollFrame.ScrollBar:GetValue() - scrollFrameHeight) == 0) then
		scrollFrame.ScrollBarScrollDownButton:Disable()
	else
		scrollFrame.ScrollBarScrollDownButton:Enable()
	end

	for i=1, GUI_Data_maxNum do
		GVAR.DataFrame.ScrollFrameButton[i]:Hide()
	end

	local j = 1
	local sliderPos = scrollFrame.offset
	for i=sliderPos+1, sliderPos+numToDisplay do
		if sliderPos < GUI_Data_Count then
			GVAR.DataFrame.ScrollFrameButton[j].NumberButton:SetText(i)
			GVAR.DataFrame.ScrollFrameButton[j].NPCButton:SetText(GUI_Data[i].npcid)
			if GUI_Data[i].exception == "zzzzz" then
				GVAR.DataFrame.ScrollFrameButton[j].ExceptionsButton:SetText("")
			else
				GVAR.DataFrame.ScrollFrameButton[j].ExceptionsButton:SetText(GUI_Data[i].exception)
			end
			if GUI_Data[i].known then
				GVAR.DataFrame.ScrollFrameButton[j].FlightPointButton:SetTextColor(0, 0.75, 0, 1)
			else
				GVAR.DataFrame.ScrollFrameButton[j].FlightPointButton:SetTextColor(1, 0, 0, 1)
			end
			GVAR.DataFrame.ScrollFrameButton[j].FlightPointButton:SetText(GUI_Data[i].name)

			if FlightMaster_PlayerNodes[GUI_Selection] and FlightMaster_PlayerNodes[GUI_Selection]["UPDATE"] then
				GVAR.DataFrame.ScrollFrameButton[j].NumberButton:SetTextColor(1, 1, 0.49, 1)
				GVAR.DataFrame.ScrollFrameButton[j].NPCButton:SetTextColor(1, 1, 1, 1)
			else
				GVAR.DataFrame.ScrollFrameButton[j].NumberButton:SetTextColor(0.3, 0.3, 0.3, 1)
				GVAR.DataFrame.ScrollFrameButton[j].NPCButton:SetTextColor(0.3, 0.3, 0.3, 1)
				GVAR.DataFrame.ScrollFrameButton[j].FlightPointButton:SetTextColor(0.3, 0.3, 0.3, 1)
			end

			GVAR.DataFrame.ScrollFrameButton[j].node = GUI_Data[i].node --print(".node =", GUI_Data[i].node)
			GVAR.DataFrame.ScrollFrameButton[j].buttonNum = i
			GVAR.DataFrame.ScrollFrameButton[j]:Show()
			j = j + 1
		end
	end
end
-- ---------------------------------------------------------------------------------------------------------------------



-- ---------------------------------------------------------------------------------------------------------------------
local function FlightMaster_TakeTaxiNode(index)
	if not flightStart then return end
	if TaxiNodeGetType(index) ~= "REACHABLE" then return end

	flightCost = TaxiNodeCost(index)
	if flightCost > GetMoney() then return end

	local tX,tY = TaxiNodePosition(index)
	local tCalc = calcID(tX,tY)
	flightDestination = tCalc
	flightCost = L["Cost"]..": "..GetMoneyString(flightCost)
	flightDuration = 0
	inTaxi = GetTime()

	FlightMaster:RegisterEvent("PLAYER_CONTROL_GAINED")
	FlightMaster:RegisterEvent("LFG_UPDATE")
	FlightMaster:RegisterEvent("PET_BATTLE_QUEUE_STATUS")
	FlightMaster:RegisterEvent("UI_ERROR_MESSAGE")
	FlightMaster:RegisterEvent("UI_SCALE_CHANGED") -- WORLDTEST test=comment
	FlightMaster:RegisterEvent("DISPLAY_SIZE_CHANGED") -- WORLDTEST test=comment

	FlightMaster:BarFrameShow()
	FlightMaster:GetNormalWorldFramePoints()
	if FlightMaster_Options.WorldView > 1 then
		FlightMaster:SetNewWorldFramePoints()
	end

	table_wipe(flightNodes)
	table_wipe(flightNone)
end

local function FlightMaster_TaxiNodeOnButtonEnter(button)
	if mapError then return end
	local index = button:GetID()
	if not index then return end
	if TaxiNodeGetType(index) ~= "REACHABLE" then return end

	local tX,tY = TaxiNodePosition(index)
	local destination = calcID(tX,tY)
	local numRoutes = GetNumRoutes(index)
	local time, taken = FlightMaster:GetFlightTime(flightStart, destination)
	local estimate = ""
	--print("+++", flightStart, destination, numRoutes, time, taken)
	
	-- 1.) other faction
	if time == 0 then
		local t = FlightMaster:GetFlightTime(flightStart, destination, 1) -- otherFaction
		if t > 0 then
			time = t
			estimate = " "..L["estimated"]
		end
	end

	-- 2.) reversed
	if time == 0 then
		local t = FlightMaster:GetFlightTime(destination, flightStart)
		if t > 0 then
			time = t
			estimate = " "..L["reversed"]
		end
	end

	-- 3.) other faction reversed
	if time == 0 then
		local t = FlightMaster:GetFlightTime(destination, flightStart, 1) -- otherFaction
		if t > 0 then
			time = t
			estimate = " "..L["reversed"]
		end
	end

	-- 4.) estimate
	local estTimeRoute = 0
	local estTimeSingle = 0
	if time == 0 then
		if numRoutes > 1 then
			TaxiNodeSetCurrent(index) -- This is magic -- it renumbers the 'hops' between index and current so the for loop below works.
			flightSteps[1] = flightStart
			--flightKnown[1] = 0

			local length = 0
			local ohcrap = false
			for i=1, numRoutes do
				local l, dx, dy = FlightMaster:LineLength(index, i)
				--print(calcID(dx,dy),dx, dy, l, i, "aa")
				length = length + l

				local isIn
				local needle = calcID(dx,dy)
				for key, value in pairs(flightNodes) do

					-- Wacky Node catcher. TaxiGetSrcX/TaxiGetSrcY/TaxiGetDestX/TaxiGetDestY vs TaxiNodePosition can differ slightly!
					if key == needle then
						isIn = key
						--print(key, dx, dy, needle, "k0")
						break
					else
						-- 1606192 number 0.60551518201828 0.19244165718555 Gadgetzan, Tanaris
						local kx = math_floor((key - 1000000) / 1000)    -- key 606
						local ky = (key - 1000000) - (kx * 1000)         -- key 192
						local nx = math_floor((needle - 1000000) / 1000) -- needle
						local ny = (needle - 1000000) - (nx * 1000)      -- needle

						if ((kx <= nx + 2) and (kx >= nx - 2) and (ky <= ny + 2) and (ky >= ny - 2)) then
							isIn = key
							--print(kx..ky, nx..ny, "(",key,"vs.",needle,")", dx,dy,"|cff11ff82-##yes-|r")
							break
						--else
							--print(kx..ky, nx..ny, "(",key,"vs.",needle,")", dx,dy,"|cffff4482-##no-|r")
						end
					end

				end

				if not isIn then
					--print(needle, dx, dy, i,"|cffffff82-crap-|r")
					ohcrap = true
					break
				end
				local tX,tY = TaxiNodePosition(flightNodes[isIn])
				flightSteps[i+1] =  calcID(tX,tY)
				--flightKnown[i+1] = 0
			end

			if not ohcrap then

				-- Get each single-hop flight between start and destination.
				local skippedSingle = 0
				local timeSingle = 0
				for x=1, #flightSteps-1 do
					local t = FlightMaster:GetFlightTime(flightSteps[x], flightSteps[x+1])
					if t == 0 then
						t = FlightMaster:GetFlightTime(flightSteps[x], flightSteps[x+1], 1) -- otherFaction
					end
					if t == 0 then
						t = FlightMaster:GetFlightTime(flightSteps[x+1], flightSteps[x])
					end
					if t == 0 then
						t = FlightMaster:GetFlightTime(flightSteps[x+1], flightSteps[x], 1) -- otherFaction
					end
					--print("$$", flightSteps[x], "(", x, ")", "->", flightSteps[x+1], "(", x+1, ")")
					if t > 0 then
						timeSingle = timeSingle + t
						--flightKnown[x] = 1
					else
						skippedSingle = skippedSingle + FlightMaster:LineLength(index, x)
					end						
				end
				--print("$#$", timeSingle, skippedSingle)

				-- Step backwards through flightSteps to find the longest interval that gives us a time.
				local skippedRoute = 0
				local timeRoute = 0
				local node = 1
				while node <= numRoutes do
					local nnode = numRoutes+1
					while nnode > node do
						local t = FlightMaster:GetFlightTime(flightSteps[node], flightSteps[nnode])
						--print("##", flightSteps[node], "(", node, ")", "->", flightSteps[nnode], "(", nnode, ")")
						if t == 0 then
							--print(nnode,node,"|",flightSteps[nnode],flightSteps[node])
							t = FlightMaster:GetFlightTime(flightSteps[nnode], flightSteps[node])
						end
						nnode = nnode - 1
						if t > 0 then
							-- We know this! Add it to the time and skip the rest of the loops.
							timeRoute = timeRoute + t
							break
						elseif nnode == node then
							-- We have no idea. Mark down that we skipped this.
							skippedRoute = skippedRoute + FlightMaster:LineLength(index, node)
							--print("skippedRoute", node, FlightMaster:LineLength(index, node))
							break
						end
					end
					node = nnode + 1
				end
				--print("*#*", timeRoute, skippedRoute)

				-- Calculate time.
				if timeSingle > 0 then
					if skippedSingle > 0 then
						local skippendSinglePercent = 100-(((length-skippedSingle)/length)*100)
						local skippendSingleTime    = timeSingle*(skippendSinglePercent/100)
						estTimeSingle = timeSingle + skippendSingleTime
						--print( "estTimeSingle (+skip):", estTimeSingle, timeSingle, length, skippedSingle, skippendSinglePercent, skippendSingleTime )
					else
						estTimeSingle = timeSingle
						--print( "estTimeSingle (-skip):", estTimeSingle, timeSingle, length, skippedSingle )
					end
					estimate = " "..L["estimated"]
				end
				if timeRoute > 0 then
					if skippedRoute > 0 then
						local skippendRoutePercent = 100-(((length-skippedRoute)/length)*100)
						local skippendRouteTime    = timeRoute*(skippendRoutePercent/100)
						estTimeRoute = timeRoute + skippendRouteTime
						--print( "estTimeRoute: (+skip)", estTimeRoute, timeRoute, length, skippedRoute, skippendRoutePercent, skippendRouteTime )
					else
						estTimeRoute = timeRoute
						--print( "estTimeRoute: (-skip)", estTimeRoute, timeRoute, length, skippedRoute )
					end
					estimate = " "..L["estimated"]
				end

			end

		end
	end
	
	if estTimeRoute > 0 and estTimeSingle > 0 then
		time = (estTimeRoute + estTimeSingle) / 2
		--print("===", time, "=", estTimeRoute, estTimeSingle)
	else
		if estTimeRoute > 0 then
			time = estTimeRoute
		elseif estTimeSingle > 0 then
			time = estTimeSingle
		end
	end
	
	BarUpdate_TimerDuration = time
	flightEstimate = estimate
	flightHops = " - "..L["Hops"]..": |cffffff82"..numRoutes.."|r"
	
	GameTooltip:AddLine(L["Time"]..": |cffffff82"..formatTime(time)..estimate.."|r"..flightHops, 1, 1, 1)
	GameTooltip:AddLine(string_format(L["Flown %s times"], "|cffffff82"..taken.."|r"), 1, 1, 1)
	GameTooltip:Show()

	--[[
	print("-START-")
	for x=1, #flightSteps-1 do
		local col = "|cffb0b0b0"
		if flightKnown[x] == 1 then
			col = "|cff00bf00"
		end
		print(x, col..flightSteps[x].."|r", "->", col..flightSteps[x+1].."|r")
	end
	print("--END--")
	--]]

	table_wipe(flightSteps)
	--table_wipe(flightKnown)
end
-- ---------------------------------------------------------------------------------------------------------------------



-- ---------------------------------------------------------------------------------------------------------------------
function FlightMaster:TaxiMapError(isError)
	mapError = isError
	if mapError then
		-- disable buttons
		TEMPLATE.DisableCheckButton(GVAR.TaxiFrameCorrectFrameLevel)
		TEMPLATE.DisableCheckButton(GVAR.TaxiFrameShowOverview)
		TEMPLATE.DisableCheckButton(GVAR.TaxiFrameShowMissing)
	else
		-- enable buttons
		TEMPLATE.EnableCheckButton(GVAR.TaxiFrameCorrectFrameLevel)
		TEMPLATE.EnableCheckButton(GVAR.TaxiFrameShowOverview)
		TEMPLATE.EnableCheckButton(GVAR.TaxiFrameShowMissing)
	end
end

function FlightMaster:Function_TAXIMAP_CLOSED()
	--print("Function_TAXIMAP_CLOSED")
	--for k,v in pairs(flightNodes) do print("flightNodes:", k,v) end
	--for k,v in pairs(flightNone) do print("flightNone:", k,v) end
	table_wipe(flightNodes)
	table_wipe(flightNone)
end

function FlightMaster:Function_TAXIMAP_OPENED()
	FlightMaster:TaxiMapError(false)

	SetMapToCurrentZone()
	local map = GetCurrentMapContinent()
	if map > 0 and map < 5 then
		map = map
	elseif map == 6 then
		map = 5
	end

	if continentID[ map ] then
		flightMap = continentID[ map ].nameID
	else
		Print("ERROR", "unknown map", locale, map)
		Print("Please contact addon author. Thanks.")
		FlightMaster:TaxiMapError(true)
		return
	end

	local npcID
	local unitGUID = UnitGUID("target")
	if unitGUID then
		npcID = tonumber( string.sub(unitGUID,6,10), 16 ) -- /script print( tonumber( string.sub(UnitGUID("target"),6,10), 16 ) )
	end
	if not npcID then
		Print("ERROR", "unknown npcID", locale, map, unitGUID, npcID)
		Print("Please contact addon author. Thanks.")
		FlightMaster:TaxiMapError(true)
		return
	end

	local numNodes = NumTaxiNodes() or 0
	--print(npcID, "#", unitGUID, "#", string.sub(unitGUID,6,10), "#", map, flightMap, numNodes, "#", NumTaxiNodes())

	if FlightMaster_NPC[playerFaction] and
	   FlightMaster_NPC[playerFaction][flightMap] and
	   not FlightMaster_NPC[playerFaction][flightMap][npcID]
	then
		local tCalc, tName
		for index = 1, numNodes do
			local tType = TaxiNodeGetType(index)
			if tType == "CURRENT" then
				local tX,tY = TaxiNodePosition(index)
				tCalc = calcID(tX,tY)
				tName = TaxiNodeName(index)
				break
			end
		end
		Print("ERROR", "unknown npc", locale, map, flightMap, npcID, tCalc, tName)
		Print("Please contact addon author. Thanks.")
		FlightMaster:TaxiMapError(true)
		return
	end

	if numNodes == 0 then
		-- known map? known npc? -> this is a new map and a new flight point -> save it
		if FlightMaster_NPC[playerFaction] and
		   FlightMaster_NPC[playerFaction][flightMap] and
		   FlightMaster_NPC[playerFaction][flightMap][npcID]
		then
			local newFlightNode = FlightMaster_NPC[playerFaction][flightMap][npcID]
			--print("newFlightNode =", newFlightNode)
			if not FlightMaster_PlayerNodes[flightMap] then FlightMaster_PlayerNodes[flightMap] = {} end
			FlightMaster_PlayerNodes[flightMap][newFlightNode] = 1
			FlightMaster_PlayerNodes[flightMap]["UPDATE"] = currentTimeDate(5)
			if GVAR.MainFrame:IsShown() then
				GUI_Selection = flightMap
				FlightMaster:UpdateData()
			end
		end
		FlightMaster:TaxiMapError(true)
		return
	end

	local taxiNodePositions = {} -- Blizz code for repositioning TaxiButton
	local numValidFlightNodes = 0 -- Blizz code for repositioning TaxiButton

	for index = 1, numNodes do
		-- save localized Nodes (all) START
		local tX,tY = TaxiNodePosition(index)
		local tCalc = calcID(tX,tY)
		local tName = TaxiNodeName(index)
		local tType = TaxiNodeGetType(index)
		if not FlightMaster_Nodes[locale]                           then FlightMaster_Nodes[locale] = {} end
		if not FlightMaster_Nodes[locale][playerFaction]            then FlightMaster_Nodes[locale][playerFaction] = {} end
		if not FlightMaster_Nodes[locale][playerFaction][flightMap] then FlightMaster_Nodes[locale][playerFaction][flightMap] = {} end
		FlightMaster_Nodes[locale][playerFaction][flightMap][tCalc] = tName
		if not FlightMaster_local_Nodes[locale]                           then FlightMaster_local_Nodes[locale] = {} end
		if not FlightMaster_local_Nodes[locale][playerFaction]            then FlightMaster_local_Nodes[locale][playerFaction] = {} end
		if not FlightMaster_local_Nodes[locale][playerFaction][flightMap] then FlightMaster_local_Nodes[locale][playerFaction][flightMap] = {} end
		FlightMaster_local_Nodes[locale][playerFaction][flightMap][tCalc] = tName
		-- eg.: deDE Alliance 1 1606192 ( 0.60551518201828 0.19244165718555 ) Gadgetzan, Tanaris
		--print(locale, tType, playerFaction, flightMap, tCalc, "(",tX,tY,")", tName)
		-- save localized Nodes (all) END

		-- --------------------------------------------------------------
		-- Blizz code for repositioning TaxiButton START -- TaxiFrame.lua
		taxiNodePositions[index] = {}
		--if tType ~= "NONE" then -- differs from TaxiFrame.lua, we check all because we show all nodes
			numValidFlightNodes = numValidFlightNodes + 1
			local currX = tX*myTAXI_MAP_WIDTH
			local currY = tY*myTAXI_MAP_HEIGHT
			taxiNodePositions[index].x = currX
			taxiNodePositions[index].y = currY
			--print("numValidFlightNodes:", numValidFlightNodes)
			if numValidFlightNodes > 1 then
				for checkNode = 1, index - 1 do
					local checkX = taxiNodePositions[checkNode].x
					local checkY = taxiNodePositions[checkNode].y
					if taxiNodePositions[checkNode].x then
						local distX = taxiNodePositions[index].x - checkX
						local distY = taxiNodePositions[index].y - checkY
						local distSq = distX*distX + distY*distY
						--print("#", index, tType, TaxiNodeName(index), tCalc, "|", distSq, "<=>", myTAXI_BUTTON_MIN_DIST * myTAXI_BUTTON_MIN_DIST)
						if distSq < myTAXI_BUTTON_MIN_DIST * myTAXI_BUTTON_MIN_DIST then
							--print("#", index, tType, TaxiNodeName(index), tCalc, "|", taxiNodePositions[index].x, taxiNodePositions[index].y )
							local scale = myTAXI_BUTTON_MIN_DIST / sqrt(distSq)
							taxiNodePositions[index].x = checkX + distX*scale
							taxiNodePositions[index].y = checkY + distY*scale
							--print("#", index, tType, TaxiNodeName(index), tCalc, "|", taxiNodePositions[index].x, taxiNodePositions[index].y )
							if not reposNode[flightMap] then reposNode[flightMap] = {} end
							reposNode[flightMap][tCalc] = calcID(taxiNodePositions[index].x/myTAXI_MAP_WIDTH, taxiNodePositions[index].y/myTAXI_MAP_HEIGHT)
							--print("#", index, tType, TaxiNodeName(index), tCalc, "|", reposNode[flightMap][tCalc] )
						end
					end
				end
			end
		--end
		-- Blizz code for repositioning TaxiButton END   -- TaxiFrame.lua
		-- --------------------------------------------------------------

		if not FlightMaster_PlayerNodes[flightMap] then FlightMaster_PlayerNodes[flightMap] = {} end
		FlightMaster_PlayerNodes[flightMap][tCalc] = 1

		if tType == "CURRENT" then
			flightStart = tCalc
			--print(tType, tCalc, npcID, tX,tY, index, tName)
		elseif tType == "NONE" then
			flightNone[tCalc] = index
		else
			flightNodes[tCalc] = index
			--print(tType, tCalc, npcID, tX,tY, index, tName)
		end
	end
	FlightMaster_PlayerNodes[flightMap]["UPDATE"] = currentTimeDate(5)

	FlightMaster:MissingNode()
	FlightMaster:CorrectFrameLevel()

	table_wipe(taxiNodePositions) -- Blizz code for repositioning TaxiButton

	GUI_Selection = flightMap
	if FlightMaster_Options.Overview then
		if GVAR.MainFrame:IsShown() then
			FlightMaster:UpdateData()
		else
			GVAR.MainFrame:Show()
		end
	else
		if GVAR.MainFrame:IsShown() then
			FlightMaster:UpdateData()
		end
	end
end
-- ---------------------------------------------------------------------------------------------------------------------



-- ---------------------------------------------------------------------------------------------------------------------
function FlightMaster:StopFlightWatch(saveData)
	FlightMaster:UnregisterEvent("PLAYER_CONTROL_GAINED")
	FlightMaster:UnregisterEvent("LFG_UPDATE")
	FlightMaster:UnregisterEvent("PET_BATTLE_QUEUE_STATUS")
	FlightMaster:UnregisterEvent("UI_ERROR_MESSAGE")
	FlightMaster:UnregisterEvent("UI_SCALE_CHANGED") -- WORLDTEST test=comment
	FlightMaster:UnregisterEvent("DISPLAY_SIZE_CHANGED") -- WORLDTEST test=comment
	FlightMaster:BarFrameHide()

	inTaxi = nil
	
	if flightDuration < 1 then
		saveData = nil
	end

	if saveData then
		local _, taken = FlightMaster:GetFlightTime(flightStart, flightDestination) -- only taken -> bonus calc in func GetFlightTime() doesn't matter
		local time = flightDuration
		--print("Stop Flight Watch", "saveData", time)

		-- Note: This can produce one (1) wrong entry if the guild level changes while in flight.
		if flightBonus then
			time = time * 1.25 -- we save the default time, this is the time a flight takes without the 25% guild bonus
		end

		FlightMaster:SaveFlightTime(flightStart, flightDestination, time, taken + 1)
		local start       = FlightMaster_local_Nodes[locale][playerFaction][flightMap][flightStart]
		local destination = FlightMaster_local_Nodes[locale][playerFaction][flightMap][flightDestination]
		Print(start, "->", destination, "-", L["Time"]..":", formatTime( flightDuration ), "-", flightCost)
	else
		Print(L["Time recording stopped."])
	end
	
	flightStart = nil
	flightDestination = nil
	flightDuration = 0
end

function FlightMaster:GetFlightTime(start, destination, otherFaction)
	if not start then return 0,0 end
	if not destination then return 0,0 end

	local faction = playerFaction
	if otherFaction then
		faction = oppositeFaction
	end

	if not FlightMaster_Data[faction] then return 0,0 end
	if not FlightMaster_Data[faction][flightMap] then return 0,0 end

	if not FlightMaster_Data[faction][flightMap][start] then
		-- eg.: 1606192 number 0.60551518201828 0.19244165718555 Gadgetzan, Tanaris
		local kx = math_floor((start - 1000000) / 1000) -- key 606
		local ky = (start - 1000000) - (kx * 1000)      -- key 192
		local newStart
		for id in pairs(FlightMaster_Data[faction][flightMap]) do
			local nx = math_floor((id - 1000000) / 1000) -- needle
			local ny = (id - 1000000) - (nx * 1000)      -- needle
			if ((kx <= nx + 2) and (kx >= nx - 2) and (ky <= ny + 2) and (ky >= ny - 2)) then
				newStart = id
				break
			end
		end
		if newStart then
			--local LocName = FlightMaster_local_Nodes[locale][playerFaction][flightMap][start] -- TEST can bug
			--print( "+++", LocName, "newStart:", newStart, "oldStart:", start, "(",faction,flightMap,")" ) -- TEST
			start = newStart
		else
			return 0,0
		end
	end

	if not FlightMaster_Data[faction][flightMap][start][destination] then
		-- eg.: 1606192 number 0.60551518201828 0.19244165718555 Gadgetzan, Tanaris
		local kx = math_floor((destination - 1000000) / 1000) -- key 606
		local ky = (destination - 1000000) - (kx * 1000)      -- key 192
		local newDestination
		for id in pairs(FlightMaster_Data[faction][flightMap][start]) do
			local nx = math_floor((id - 1000000) / 1000) -- needle
			local ny = (id - 1000000) - (nx * 1000)      -- needle
			if ((kx <= nx + 2) and (kx >= nx - 2) and (ky <= ny + 2) and (ky >= ny - 2)) then
				newDestination = id
				break
			end
		end
		if newDestination then
			--local LocName = FlightMaster_local_Nodes[locale][playerFaction][flightMap][destination] -- TEST can bug
			--print( "===", LocName, "newDestination:", newDestination, "oldDestination:", destination, "(",faction,flightMap,start,")" ) -- TEST
			destination = newDestination
		else
			return 0,0
		end
	end

	local s = FlightMaster_Data[faction][flightMap][start][destination]
	if type(s) ~= "string" then return 0,0 end

	local time, taken = string_match(s, ":(%d+):(%d+):")
	time = tonumber(time) or 0

	if flightBonus then
		--print("time without bonus:", time)
		time = time / 1.25
		--print("time with bonus:", time)
	end

	return time, (tonumber(taken) or 0)
end

function FlightMaster:SaveFlightTime(start, destination, time, taken)
	if not start then return end
	if not destination then return end

	local ntime = math_floor(time) or 0
	local ntaken = taken or 0

	if not FlightMaster_Data[playerFaction]                                then FlightMaster_Data[playerFaction] = {} end
	if not FlightMaster_Data[playerFaction][flightMap]                     then FlightMaster_Data[playerFaction][flightMap] = {} end
	if not FlightMaster_Data[playerFaction][flightMap][start]              then FlightMaster_Data[playerFaction][flightMap][start] = {} end
	if not FlightMaster_Data[playerFaction][flightMap][start][destination] then FlightMaster_Data[playerFaction][flightMap][start][destination] = {} end
	FlightMaster_Data[playerFaction][flightMap][start][destination] = ":"..ntime..":"..ntaken..":"
	--print("SaveFlightTime", FlightMaster_Data[playerFaction][flightMap][start][destination])
	--print("SaveFlightTime", "ntime:", ntime, "ntaken:", ntaken)
end

function FlightMaster:LineLength(index, i)
	-- start node
	local sx = TaxiGetSrcX(index, i)
	local sy = TaxiGetSrcY(index, i)
	-- end node
	local ex = TaxiGetDestX(index, i)
	local ey = TaxiGetDestY(index, i)
	-- determine dimensions
	local dx,dy = ex - sx, ey - sy
	-- normalize direction if necessary
	if (dx < 0) then
		dx,dy = -dx,-dy
	end
	-- return length and endpoint coords
	return sqrt((dx * dx) + (dy * dy)), ex, ey
end
-- ---------------------------------------------------------------------------------------------------------------------



-- ---------------------------------------------------------------------------------------------------------------------
function FlightMaster:Check4Bonus()
	-- GetGuildLevel() can return 25 even if the player is not in a guild at login.
	-- Don't know if this was a rare Beta bug [5.0.4.15929 Beta with a premade char] or whatever.
	-- Combined with IsInGuild() it always works as intended.
	if IsInGuild() then
		local guildLevel = GetGuildLevel() or 0
		if guildLevel >= 21 then -- Guild Perk Level 21 provides a 25% bonus on flight time
			flightBonus = true
		else
			flightBonus = nil
		end
	else
		flightBonus = nil
	end
	--print("flightBonus:", flightBonus, "#", GetGuildLevel(), "#", IsInGuild())
end
-- ---------------------------------------------------------------------------------------------------------------------

-- ---------------------------------------------------------------------------------------------------------------------
function FlightMaster:CheckFaction()
	playerFaction = UnitFactionGroup("player")
	if playerFaction == "Horde" then
		oppositeFaction = "Alliance"
		FlightMaster:UnregisterEvent("NEUTRAL_FACTION_SELECT_RESULT")
	elseif playerFaction == "Alliance" then
		oppositeFaction = "Horde"
		FlightMaster:UnregisterEvent("NEUTRAL_FACTION_SELECT_RESULT")
	elseif playerFaction == "Neutral" then
		playerFaction = "Alliance" -- Dummy
		oppositeFaction = "Horde" -- Dummy
		FlightMaster:RegisterEvent("NEUTRAL_FACTION_SELECT_RESULT")
	else
		Print("ERROR", "unknown faction", locale, playerFaction)
		Print("Program terminated.")
		return
	end
end
-- ---------------------------------------------------------------------------------------------------------------------

-- ---------------------------------------------------------------------------------------------------------------------
local function OnEvent(self, event, ...)
	--print(event, ...)
	if event == "TAXIMAP_OPENED" then
		FlightMaster:Function_TAXIMAP_OPENED()
	elseif event == "TAXIMAP_CLOSED" then
		FlightMaster:Function_TAXIMAP_CLOSED()

	elseif event == "PLAYER_ENTERING_WORLD" then
		inWorld = true
	elseif event == "PLAYER_LEAVING_WORLD" then
		inWorld = nil

	elseif event == "LFG_UPDATE" then
		if not inTaxi then return end
		local proposalExists, _, _, _, _, _, _, hasResponded = GetLFGProposal()
		--print(event, "GetLFGProposal():", proposalExists, hasResponded)
		if proposalExists and hasResponded then
			FlightMaster:StopFlightWatch()
		end

	elseif event == "PET_BATTLE_QUEUE_STATUS" then
		if not inTaxi then return end
		local pbStatus = C_PetBattles.GetPVPMatchmakingInfo()
		--print("pbStatus: ", pbStatus)
		if pbStatus == "entry" then
			FlightMaster:StopFlightWatch()
		end

	elseif event == "PLAYER_CONTROL_GAINED" then
		if not inWorld then return end
		if not inTaxi then return end
		flightDuration = math.ceil(GetTime() - inTaxi - 0.5)
		FlightMaster:StopFlightWatch(true)

	elseif event == "UI_ERROR_MESSAGE" then
		if not inTaxi then return end
		local msg = ...
		if TaxiErrors[msg] then
			--print("UI_ERROR_MESSAGE:", msg)
			FlightMaster:StopFlightWatch()
		end

	elseif event == "UI_INFO_MESSAGE" then
		local msg = ...
		if msg == ERR_NEWTAXIPATH then
			--print("UI_INFO_MESSAGE:", msg)
			FlightMaster:Function_TAXIMAP_OPENED()
		end

	elseif event == "GUILD_PERK_UPDATE" then
		FlightMaster:Check4Bonus()

	elseif event == "NEUTRAL_FACTION_SELECT_RESULT" then
		FlightMaster:CheckFaction()

	elseif event == "UI_SCALE_CHANGED" or
	       event == "DISPLAY_SIZE_CHANGED" then
		if not inWorld then return end
		if not inTaxi then return end -- WORLDTEST test=comment
		if not GVAR.BarFrame:IsShown() then return end
		barEffScale = GVAR.BarFrame:GetEffectiveScale()
		barScale = GVAR.BarFrame:GetScale()
		if FlightMaster_Options.WorldView > 1 then
			GVAR.BarFrame:SetScale( UIParent:GetScale() )
		end
		--print(GetTime(), UIParent:IsShown(), "#", UIParent:GetScale(), "#", barEffScale, "|", barScale, "#", event)

	elseif event == "PLAYER_LOGIN" then
		FlightMaster:CheckFaction()
		FlightMaster:InitOptions()
		FlightMaster:CreateInterfaceOptions()
		FlightMaster:LDBcheck()
		FlightMaster:CreateMinimapButton()
		FlightMaster:CreateMainFrame()
		FlightMaster:InitializeFrame()
		FlightMaster:Check4Bonus()
		table.insert(UISpecialFrames, "FlightMaster_MainFrame")

		hooksecurefunc("TakeTaxiNode", FlightMaster_TakeTaxiNode)
		hooksecurefunc("TaxiNodeOnButtonEnter", FlightMaster_TaxiNodeOnButtonEnter)

		-- actions that stops flight-time recording
		-- Summon -OK-
		hooksecurefunc("ConfirmSummon", function()
			if not inTaxi then return end
			FlightMaster:StopFlightWatch()
		end)
		-- Battlefield -OK-
		hooksecurefunc("AcceptBattlefieldPort", function(battleID, joined)
			if not inTaxi then return end
			if joined and joined == 1 then
				FlightMaster:StopFlightWatch()
			end
		end)
		-- World PvP -OK-
		hooksecurefunc("BattlefieldMgrEntryInviteResponse", function(data, joined)
			if not inTaxi then return end
			if joined and joined == 1 then
				FlightMaster:StopFlightWatch()
			end
		end)

		FlightMaster:UnregisterEvent("PLAYER_LOGIN")
	end
end
-- ---------------------------------------------------------------------------------------------------------------------

--FlightMaster:RegisterEvent("UI_SCALE_CHANGED") -- WORLDTEST test=uncomment
--FlightMaster:RegisterEvent("DISPLAY_SIZE_CHANGED") -- WORLDTEST test=uncomment
FlightMaster:RegisterEvent("PLAYER_ENTERING_WORLD")
FlightMaster:RegisterEvent("PLAYER_LEAVING_WORLD")
FlightMaster:RegisterEvent("TAXIMAP_OPENED")
FlightMaster:RegisterEvent("TAXIMAP_CLOSED")
FlightMaster:RegisterEvent("UI_INFO_MESSAGE")
FlightMaster:RegisterEvent("GUILD_PERK_UPDATE")
FlightMaster:RegisterEvent("PLAYER_LOGIN")
FlightMaster:SetScript("OnEvent", OnEvent)